# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['extract_circuit', 'extract_simple', 'graph_to_swaps', 'extract_clifford_normal_form',
           'lookahead_extract_base', 'lookahead_full', 'lookahead_fast', 'lookahead_extract']

from fractions import Fraction
import itertools

from .utils import EdgeType, VertexType, toggle_edge
from .linalg import Mat2, Z2
from .simplify import id_simp, tcount,full_reduce, is_graph_like
from .rewrite_rules.rules import apply_rule, pivot, match_spider_parallel, spider
from .circuit import Circuit
from .circuit.gates import Gate, ParityPhase, CNOT, HAD, ZPhase, XPhase, CZ, XCX, SWAP, InitAncilla

from .graph.base import BaseGraph, VT, ET

from typing import List, Optional, Tuple, Dict, Set, Union, Iterator


def bi_adj(g: BaseGraph[VT,ET], vs:List[VT], ws:List[VT]) -> Mat2:
    """Construct a biadjacency matrix between the supplied list of vertices
    ``vs`` and ``ws``."""
    return Mat2([[1 if g.connected(v,w) else 0 for v in vs] for w in ws])

def connectivity_from_biadj(
        g: BaseGraph[VT,ET], 
        m: Mat2, 
        left:List[VT], 
        right: List[VT], 
        edgetype:EdgeType=EdgeType.HADAMARD):
    """Replace the connectivity in ``g`` between the vertices in ``left`` and ``right``
    by the biadjacency matrix ``m``. The edges will be of type ``edgetype``."""
    for i in range(len(right)):
        for j in range(len(left)):
            if m.data[i][j] and not g.connected(right[i],left[j]):
                g.add_edge((right[i],left[j]),edgetype)
            elif not m.data[i][j] and g.connected(right[i],left[j]):
                g.remove_edge(g.edge(right[i],left[j]))

def streaming_extract(
        g:BaseGraph[VT,ET], 
        optimize_czs:bool=True, 
        optimize_cnots:int=2, 
        quiet:bool=True
        ) -> Circuit:
    print("This function is deprecated. Call extract_circuit() instead.")
    return extract_circuit(g, optimize_czs, optimize_cnots, quiet)

def permutation_as_swaps(perm:Dict[int,int]) -> List[Tuple[int,int]]:
    """Returns a series of swaps that realises the given permutation. 

    Args:
        perm: A dictionary where both keys and values take values in 0,1,...,n."""
    swaps = []
    l = [perm[i] for i in range(len(perm))]
    pinv = {v:k for k,v in perm.items()}
    linv = [pinv[i] for i in range(len(pinv))]
    for i in range(len(perm)):
        if l[i] == i: continue
        t1 = l[i]
        t2 = linv[i]
        swaps.append((i,t2))
        #l[i] = i
        #linv[i] = i
        l[t2] = t1
        linv[t1] = t2
    return swaps


def column_optimal_swap(m: Mat2) -> Dict[int,int]:
    """Given a matrix m, tries to find a permutation of the columns such that
    there are as many ones on the diagonal as possible. 
    This reduces the number of row operations needed to do Gaussian elimination.
    """
    r, c = m.rows(), m.cols()
    connections:  Dict[int,Set[int]] = {i: set() for i in range(r)}
    connectionsr: Dict[int,Set[int]] = {j: set() for j in range(c)}

    for i in range(r):
            for j in range(c):
                if m.data[i][j]: 
                    connections[i].add(j)
                    connectionsr[j].add(i)

    target = _find_targets(connections, connectionsr)
    if target is None: target = dict()
    #target = {v:k for k,v in target.items()}
    left = list(set(range(c)).difference(target.values()))
    right = list(set(range(c)).difference(target.keys()))
    for i in range(len(left)):
        target[right[i]] = left[i]
    return target

def _find_targets(
        conn: Dict[int,Set[int]],
        connr: Dict[int,Set[int]],
        target:Dict[int,int]={}
        ) -> Optional[Dict[int,int]]:
    """Helper function for :func:`column_optimal_swap`.
    Recursively makes a choice for a permutation that places additional ones on the diagonal.
    Backtracks when it gets stuck in an unfavorable configuration."""
    target = target.copy()
    r = len(conn)
    c = len(connr)

    claimedcols = set(target.keys())
    claimedrows = set(target.values())

    while True:
        min_index = -1
        min_options = set(range(1000))
        for i in range(r):
            if i in claimedrows: continue
            s = conn[i] - claimedcols # The free columns
            if len(s) == 1:
                j = s.pop()
                target[j] = i
                claimedcols.add(j)
                claimedrows.add(i)
                break
            if len(s) == 0: return None # contradiction
            found_col = False
            for j in s:
                t = connr[j] - claimedrows
                if len(t) == 1: # j can only be connected to i
                    target[j] = i
                    claimedcols.add(j)
                    claimedrows.add(i)
                    found_col = True
                    break
            if found_col: break
            if len(s) < len(min_options):
                min_index = i
                min_options = s
        else: # Didn't find any forced choices
            if not (conn.keys() - claimedrows): # we are done
                return target
            if min_index == -1: raise ValueError("This shouldn't happen ever")
            # Start depth-first search
            tgt = target.copy()
            #print("backtracking on", min_index)
            for i2 in min_options:
                #print("trying option", i2)
                tgt[i2] = min_index
                new_target = _find_targets(conn, connr, tgt)
                if new_target: return new_target
            #print("Unsuccessful")
            return target


def xor_rows(l1: List[Z2], l2: List[Z2]) -> List[Z2]:
    return [0 if l1[i]==l2[i] else 1 for i in range(len(l1))]


def find_minimal_sums(m: Mat2, reversed_search=False) -> Optional[Tuple[int, ...]]:
    """Returns a list of rows in m that can be added together to reduce one of the rows so that
    it only contains a single 1. Used in :func:`greedy_reduction`"""
    r = m.rows()
    d = m.data
    if any(sum(r) == 1 for r in d):
        return tuple()
    combs:  Dict[Tuple[int, ...], List[Z2]] = {(i,): d[i] for i in range(r)}
    combs2: Dict[Tuple[int, ...], List[Z2]] = {}
    iterations = 0
    while True:
        combs2 = {}
        for index, l in combs.items():
            max_index: int = max(index)
            rr: range = range(max_index + 1, r) if not reversed_search else range(r - 1, max_index, -1)
            for k in rr:
                # Unrolled xor_rows(combs[index],d[k])
                row: List[Z2] = [0 if v1 == v2 else 1 for v1, v2 in zip(combs[index], d[k])]
                # row = xor_rows(combs[index],d[k])
                if sum(row) == 1:
                    return (*index, k)
                combs2[(*index, k)] = row
                iterations += 1
            if iterations > 100000:
                return None
        if not combs2:
            return None
            # raise ValueError("Irreducible input has been given")
        combs = combs2


def greedy_reduction(m: Mat2) -> Optional[List[Tuple[int, int]]]:
    """Returns a list of tuples (r1,r2) that specify which row should be added to which other row
    in order to reduce one row of m to only contain a single 1. 
    Used in :func:`extract_circuit` and :func:`lookahead_extract_base`"""
    indicest = find_minimal_sums(m)
    if indicest is None: return indicest
    indices = list(indicest)
    rows = {i:m.data[i] for i in indices}
    weights: Dict[int,int] = {i: sum(r) for i,r in rows.items()}
    result = []
    while len(indices)>1:
        best = (-1,-1)
        reduction = -10000
        for i in indices:
            for j in indices:
                if j <= i: continue
                w = sum(xor_rows(rows[i],rows[j]))
                if weights[i] - w > reduction:
                    best = (j,i) # "Add row j to i"
                    reduction = weights[i] - w
                if weights[j] - w > reduction:
                    best = (i,j)
                    reduction = weights[j] - w
        result.append(best)
        control, target = best
        rows[target] = xor_rows(rows[control],rows[target])
        weights[target] = weights[target] - reduction
        indices.remove(control)
    return result


def flat_indices(m: Mat2, indices: List[int]) -> Tuple[List[Tuple[int, int]], int]:
    """Given a matrix and a list of row indices that have to be added together,
    returns a list of row operations and the index of the row that would end up with
    the sum of the input rows.

    When transformed into CNOTs, the depth of the circuit is log_2(len(indices)).

    If the list of indices is empty, returns ([], -1)"""
    if len(indices) == 0:
        return [], -1
    rows = {i: m.data[i] for i in indices}
    weights: Dict[int,int] = {i: sum(r) for i, r in rows.items()}
    result = []
    next_indices = []
    while len(indices) > 1:
        best = (-1, -1)
        reduction = -10000
        for i in indices:
            for j in indices:
                if j <= i: continue
                w = sum(xor_rows(rows[i], rows[j]))
                if weights[i] - w > reduction:
                    best = (j, i)  # "Add row j to i"
                    reduction = weights[i] - w
                if weights[j] - w > reduction:
                    best = (i, j)
                    reduction = weights[j] - w
        result.append(best)
        control, target = best
        rows[target] = xor_rows(rows[control], rows[target])
        weights[target] = weights[target] - reduction
        indices.remove(control)
        indices.remove(target)
        next_indices.append(target)
        if len(indices) <= 1:
            if len(indices) == 1:
                next_indices.append(indices[0])
            indices = next_indices
            next_indices = []
    return result, indices[0]


def greedy_reduction_flat(m: Mat2) -> Optional[List[Tuple[int, int]]]:
    """Returns a list of tuples (r1,r2) that specify which row should be added to which other row
    in order to reduce one row of m to only contain a single 1.
    In contrast to :func:`greedy_reduction`, it performs the brute-force search starting with the
    highest indices, and places the row operations in such a way that the resulting depth is log_2
    of the number of rows that have to be added together.
    Used in :func:`lookahead_extract_base`"""
    indicest = find_minimal_sums(m, True)
    if indicest is None: return indicest
    return flat_indices(m, list(indicest))[0]


def find_2_minimal_sums(m: Mat2) -> Optional[Tuple[Tuple[int, ...], Tuple[int, ...]]]:
    """Returns two lists of rows in m that can be added together to reduce two of the rows so that
    they only contains a single 1. Used in :func:`greedy_two_reduction`"""
    r = m.rows()
    d = m.data
    combs:  Dict[Tuple[int, ...], List[Z2]] = {(i,): d[i] for i in range(r)}
    combs2: Dict[Tuple[int, ...], List[Z2]]
    sum1: Optional[Tuple[int, ...]] = None
    iterations = 0
    while True:
        combs2 = {}
        for index, l in combs.items():
            for k in range(max(index)+1, r):
                row: List[Z2] = [0 if v1 == v2 else 1 for v1, v2 in zip(combs[index], d[k])]
                if sum(row) == 1:
                    if sum1 is None:
                        sum1 = (*index, k)
                    else:
                        return sum1, (*index, k)
                combs2[(*index, k)] = row
                iterations += 1
            if iterations > 100000:
                return None
        if not combs2:
            return None
        combs = combs2


def greedy_two_reduction(m: Mat2) -> Optional[List[Tuple[int, int]]]:
    """Returns a list of tuples (r1,r2) that specify which row should be added to which other row
    in order to reduce two rows (instead of one as the other greedy reductions) of m to only contain a single 1.
    Used in :func:`lookahead_extract`"""
    indicest = find_2_minimal_sums(m)
    if indicest is None:
        return indicest
    s1 = set(indicest[0])
    s2 = set(indicest[1])

    if s2.issubset(s1):
        s3 = s2
        s2 = s1
        s1 = s3
    if s1.issubset(s2):
        res0, r0 = flat_indices(m, list(s1))
        res1, r1 = flat_indices(m, list(s2.difference(s1)))
        res0.extend(res1)
        if r1 == -1:
            # Should never happen
            print("Got the same sets: {}, {}".format(s1, s2))
            return None
        res0.append((r0, r1))
        return res0

    res0, r0 = flat_indices(m, list(s1.intersection(s2)))
    res1, r1 = flat_indices(m, list(s1.difference(s2)))
    res2, r2 = flat_indices(m, list(s2.difference(s1)))
    res0.extend(res1)
    res0.extend(res2)
    if r0 != -1:
        res0.append((r0, r1))
        res0.append((r0, r2))
    return res0


# O(N^3)
def max_overlap(cz_matrix: Mat2) -> Tuple[Tuple[int,int],List[int]]:
    """Given an adjacency matrix of qubit connectivity of a CZ circuit, returns:
    a) the rows which have the maximum inner product
    b) the list of common qubits between these rows.
    Used in :func:`extract_circuit` to more optimally place CZ gates. 
    """
    N = len(cz_matrix.data[0])

    max_inner_product = 0
    final_common_qbs = list()
    overlapping_rows = (-1,-1)
    for i in range(N):
        for j in range(i+1,N):
            inner_product = 0
            i_czs = 0
            j_czs = 0
            common_qbs = list()
            for k in range(N):
                i_czs += cz_matrix.data[i][k]
                j_czs += cz_matrix.data[j][k]

                if cz_matrix.data[i][k]!=0 and cz_matrix.data[j][k]!=0:
                    inner_product+=1
                    common_qbs.append(k)

            if inner_product > max_inner_product:
                max_inner_product = inner_product
                if i_czs < j_czs:
                    overlapping_rows = (j,i)
                else:
                    overlapping_rows = (i,j)
                final_common_qbs = common_qbs
    return (overlapping_rows,final_common_qbs)

def filter_duplicate_cnots(cnots: List[CNOT]) -> List[CNOT]:
    """Cancels adjacent CNOT gates in a list of CNOT gates."""
    from .optimize import basic_optimization
    if not cnots: return cnots  # Nothing to do if the list is empty
    qubits = max([max(cnot.control,cnot.target) for cnot in cnots]) + 1
    c = Circuit(qubits)
    c.gates = cnots.copy() # type: ignore
    c = basic_optimization(c,do_swaps=False)
    return c.gates # type: ignore


def remove_extra_cnots(cnots_to_apply: List[CNOT], mat: Mat2) -> List[CNOT]:
    """Remove redundant CNOTs and return a list of only the necessary ones"""
    cnots = cnots_to_apply.copy()
    m2 = mat.copy()
    for cnot in cnots:
        m2.row_add(cnot.target, cnot.control)
    extractable = set()
    for i, row in enumerate(m2.data):
        if sum(row) == 1:
            extractable.add(i)
    # We now know which vertices are extractable, and hence the CNOTs on qubits that do not involve
    # these vertices aren't necessary.
    # So first, we get rid of all the CNOTs that happen in the Gaussian elimination after
    # all the extractable vertices have become extractable
    m2 = mat.copy()
    for count, cnot in enumerate(cnots):
        if sum(1 for row in m2.data if sum(row) == 1) == len(extractable):  # extractable rows equal to maximum
            cnots = cnots[:count]  # So we do not need the remainder of the CNOTs
            break
        m2.row_add(cnot.target, cnot.control)
    # We now recalculate which vertices were extractable, because the deleted cnots
    # might have acted to swap this vertex around some.
    extractable = set()
    for i, row in enumerate(m2.data):
        if sum(row) == 1:
            extractable.add(i)
    # And now we try to get rid of some more CNOTs, that can be commuted to the end of the CNOT circuit
    # without changing extractability.
    necessary_cnots = []
    # 'A' stands for "blocked for All". 'R' for "blocked for Red", 'G' for "blocked for Green".
    blocked = {i: 'A' for i in extractable}
    for cnot in reversed(cnots):
        if cnot.target not in blocked and cnot.control not in blocked:
            continue  # CNOT not needed
        should_add = False
        if cnot.target in blocked and blocked[cnot.target] != 'R':
            should_add = True
            blocked[cnot.target] = 'A'
        if cnot.control in blocked and blocked[cnot.control] != 'G':
            should_add = True
            blocked[cnot.control] = 'A'
        if cnot.control in extractable:
            should_add = True
        if cnot.target in extractable:
            should_add = True
        if not should_add:
            continue
        necessary_cnots.append(cnot)
        if cnot.control not in blocked:
            blocked[cnot.control] = 'G'  # 'G' stands for Green
        if cnot.target not in blocked:
            blocked[cnot.target] = 'R'  # 'R' stands for Red

    return list(reversed(necessary_cnots))


def apply_cnots(g: BaseGraph[VT, ET], c: Circuit, frontier: List[VT], qubit_map: Dict[VT, int],
                cnots: List[CNOT], m: Mat2, neighbors: List[VT]) -> int:
    """Adds the list of CNOTs to the circuit, modifying the graph, frontier, and qubit map as needed.
    Returns the number of vertices that end up being extracted"""
    if len(cnots) > 0:
        cnots2 = cnots
        cnots = []
        for cnot in cnots2:
            m.row_add(cnot.target, cnot.control)
            cnots.append(CNOT(qubit_map[frontier[cnot.control]], qubit_map[frontier[cnot.target]]))
        connectivity_from_biadj(g, m, neighbors, frontier)

    good_verts = dict()
    for i, row in enumerate(m.data):
        if sum(row) == 1:
            v = frontier[i]
            w = neighbors[[j for j in range(len(row)) if row[j]][0]]
            good_verts[v] = w
    if not good_verts:
        raise Exception("No extractable vertex found. Something went wrong")
    hads = []
    outputs = g.outputs()
    for v, w in good_verts.items():  # Update frontier vertices
        hads.append(qubit_map[v])
        # c.add_gate("HAD",qubit_map[v])
        qubit_map[w] = qubit_map[v]
        b = [o for o in g.neighbors(v) if o in outputs][0]
        g.remove_vertex(v)
        g.add_edge((w, b))
        frontier.remove(v)
        frontier.append(w)

    for cnot in cnots:
        c.add_gate(cnot)
    for h in hads:
        c.add_gate("HAD", h)

    return len(good_verts)


def clean_frontier(g: BaseGraph[VT, ET], c: Circuit, frontier: List[VT],
                   qubit_map: Dict[VT, int], optimize_czs: bool = True) -> int:
    """Remove single qubit gates from the frontier and any CZs between the vertices in the frontier
    Returns the number of CZs saved if `optimize_czs` is True; otherwise returns 0"""
    phases = g.phases()
    czs_saved = 0
    outputs = g.outputs()
    for v in frontier:  # First removing single qubit gates
        q = qubit_map[v]
        b = [w for w in g.neighbors(v) if w in outputs][0]
        e = g.edge(v, b)
        if g.edge_type(e) == EdgeType.HADAMARD:
            c.add_gate("HAD", q)
            g.set_edge_type(e, EdgeType.SIMPLE)
        if phases[v]:
            c.add_gate("ZPhase", q, phases[v])
            g.set_phase(v, 0)
    # And now on to CZ gates
    cz_mat = Mat2([[0 for i in range(len(outputs))] for j in range(len(outputs))])
    for v in frontier:
        for w in list(g.neighbors(v)):
            if w in frontier:
                cz_mat.data[qubit_map[v]][qubit_map[w]] = 1
                cz_mat.data[qubit_map[w]][qubit_map[v]] = 1
                g.remove_edge(g.edge(v, w))

    if optimize_czs:
        overlap_data = max_overlap(cz_mat)
        while len(overlap_data[1]) > 2:  # there are enough common qubits to be worth optimizing
            i, j = overlap_data[0][0], overlap_data[0][1]
            czs_saved += len(overlap_data[1]) - 2
            c.add_gate("CNOT", i, j)
            for qb in overlap_data[1]:
                c.add_gate("CZ", j, qb)
                cz_mat.data[i][qb] = 0
                cz_mat.data[j][qb] = 0
                cz_mat.data[qb][i] = 0
                cz_mat.data[qb][j] = 0
            c.add_gate("CNOT", i, j)
            overlap_data = max_overlap(cz_mat)

    for i in range(len(outputs)):
        for j in range(i + 1, len(outputs)):
            if cz_mat.data[i][j] == 1:
                c.add_gate("CZ", i, j)

    return czs_saved


def neighbors_of_frontier(g: BaseGraph[VT, ET], frontier: List[VT]) -> Set[VT]:
    """Returns the set of neighbors of the frontier. When collecting the vertices, it also checks if the vertices
    of the frontier are connected correctly to the inputs.
    If a frontier vertex is only connected to an input, it is removed from the frontier.
    If a frontier vertex is connected to an input and some other vertices, it is disconnected from the input via a new
    spider."""
    qs = g.qubits()
    rs = g.rows()
    neighbor_set = set()
    inputs = g.inputs()
    outputs = g.outputs()
    for v in frontier.copy():
        d = [w for w in g.neighbors(v) if w not in outputs]
        if any(w in inputs for w in d):  # frontier vertex v is connected to an input
            if len(d) == 1:  # Only connected to input, remove from frontier
                frontier.remove(v)
                continue
            # We disconnect v from the input b via a new spider
            b = [w for w in d if w in inputs][0]
            q = qs[b]
            r = rs[b]
            w = g.add_vertex(VertexType.Z, q, r + 1)
            e = g.edge(v, b)
            et = g.edge_type(e)
            g.remove_edge(e)
            g.add_edge((v, w), EdgeType.HADAMARD)
            g.add_edge((w, b), toggle_edge(et))
            d.remove(b)
            d.append(w)
        neighbor_set.update(d)
    return neighbor_set


def remove_gadget(g: BaseGraph[VT, ET], frontier: List[VT], qubit_map: Dict[VT, int],
                  neighbor_set: Set[VT], gadgets: Dict[VT, VT]) -> bool:
    """Removes a gadget that is attached to a frontier vertex. Returns True if such gadget was found, False otherwise"""
    removed_gadget = False
    outputs = g.outputs()
    for w in neighbor_set:
        if w not in gadgets: continue
        for v in g.neighbors(w):
            if v in frontier:
                apply_rule(g, pivot, [((w, v), ([], [o for o in g.neighbors(v) if o in outputs]))])  # type: ignore
                frontier.remove(v)
                del gadgets[w]
                frontier.append(w)
                qubit_map[w] = qubit_map[v]
                removed_gadget = True
                break
    return removed_gadget


def extract_circuit(
        g: BaseGraph[VT, ET],
        optimize_czs: bool = True,
        optimize_cnots: int = 2,
        up_to_perm: bool = False,
        quiet: bool = True
        ) -> Circuit:
    """Given a graph put into semi-normal form by :func:`~pyzx.simplify.full_reduce`, 
    it extracts its equivalent set of gates into an instance of :class:`~pyzx.circuit.Circuit`.
    This function implements a more optimized version of the algorithm described in
    `There and back again: A circuit extraction tale <https://arxiv.org/abs/2003.01664>`_

    Args:
        g: The ZX-diagram graph to be extracted into a Circuit.
        optimize_czs: Whether to try to optimize the CZ-subcircuits by exploiting overlap between the CZ gates
        optimize_cnots: (0,1,2,3) Level of CNOT optimization to apply.
        up_to_perm: If true, returns a circuit that is equivalent to the given graph up to a permutation of the inputs.
        quiet: Whether to print detailed output of the extraction process.

    Warning:
        Note that this function changes the graph `g` in place. 
        In particular, if the extraction fails, the modified `g` shows 
        how far the extraction got. If you want to keep the original `g`
        then input `g.copy()` into `extract_circuit`.
    """

    gadgets = {}
    inputs = g.inputs()
    outputs = g.outputs()

    c = Circuit(len(outputs))

    if not is_graph_like(g):
        raise ValueError("Input graph is not graph-like. Try running full_reduce first")

    for v in g.vertices():
        if g.vertex_degree(v) == 1 and v not in inputs and v not in outputs:
            n = list(g.neighbors(v))[0]
            gadgets[n] = v

    qubit_map: Dict[VT,int] = dict()
    frontier = []
    for i, o in enumerate(outputs):
        v = list(g.neighbors(o))[0]
        if v in inputs:
            continue
        frontier.append(v)
        qubit_map[v] = i

    czs_saved = 0
    q: Union[float, int]
    
    while True:
        # preprocessing
        czs_saved += clean_frontier(g, c, frontier, qubit_map, optimize_czs)
        
        # Now we can proceed with the actual extraction
        # First make sure that frontier is connected in correct way to inputs
        neighbor_set = neighbors_of_frontier(g, frontier)
        
        if not frontier:
            break  # No more vertices to be processed. We are done.
        
        # First we check if there is a phase gadget in the way
        if remove_gadget(g, frontier, qubit_map, neighbor_set, gadgets):
            # There was a gadget in the way. Go back to the top
            continue
            
        neighbors = list(neighbor_set)
        m = bi_adj(g, neighbors, frontier)
        if all(sum(row) != 1 for row in m.data):  # No easy vertex
            if optimize_cnots > 1:
                greedy_operations = greedy_reduction(m)
            else:
                greedy_operations = None

            if greedy_operations is not None:
                greedy = [CNOT(target, control) for control, target in greedy_operations]
                if (len(greedy) == 1 or optimize_cnots < 3) and not quiet:
                    print("Found greedy reduction with", len(greedy), "CNOT")
                cnots = greedy

            if greedy_operations is None or (optimize_cnots == 3 and len(greedy) > 1):
                perm = column_optimal_swap(m)
                perm = {v: k for k, v in perm.items()}
                neighbors2 = [neighbors[perm[i]] for i in range(len(neighbors))]
                m2 = bi_adj(g, neighbors2, frontier)
                if optimize_cnots > 0:
                    cnots = m2.to_cnots(optimize=True)
                else:
                    cnots = m2.to_cnots(optimize=False)
                # Since the matrix is not square, the algorithm sometimes introduces duplicates
                cnots = filter_duplicate_cnots(cnots)

                if greedy_operations is not None:
                    m3 = m2.copy()
                    for cnot in cnots:
                        m3.row_add(cnot.target,cnot.control)
                    reductions = sum(1 for row in m3.data if sum(row) == 1)
                    if greedy and (len(cnots)/reductions > len(greedy)-0.1):
                        if not quiet: print("Found greedy reduction with", len(greedy), "CNOTs")
                        cnots = greedy
                    else:
                        neighbors = neighbors2
                        m = m2
                        if not quiet: print("Gaussian elimination with", len(cnots), "CNOTs")
            # We now have a set of CNOTs that suffice to extract at least one vertex.
        else:
            if not quiet: print("Simple vertex")
            cnots = []

        extracted = apply_cnots(g, c, frontier, qubit_map, cnots, m, neighbors)
        if not quiet: print("Vertices extracted:", extracted)
            
    if optimize_czs:
        if not quiet: print("CZ gates saved:", czs_saved)
    # Outside of loop. Finish up the permutation
    id_simp(g, quiet=True)  # Now the graph should only contain inputs and outputs
    # Since we were extracting from right to left, we reverse the order of the gates
    c.gates = list(reversed(c.gates))
    return graph_to_swaps(g, up_to_perm) + c


def extract_simple(g: BaseGraph[VT, ET], up_to_perm: bool = True) -> Circuit:
    """A simplified circuit extractor that works on graphs with a causal flow (e.g. graphs arising
    from circuits via spider fusion).

    Args:
        g: The graph to extract
        up_to_perm: If true, returns a circuit that is equivalent to the given graph up to a permutation of the inputs.
    """
    
    progress = True
    outputs = g.outputs()
    circ = Circuit(len(outputs))
    while progress:
        progress = False
        
        for q, o in enumerate(outputs):
            if g.vertex_degree(o) != 1:
                raise ValueError("Bad output degree")
            v = list(g.neighbors(o))[0]
            e = g.edge(o, v)
            
            if g.edge_type(e) == EdgeType.HADAMARD:
                progress = True
                circ.prepend_gate(HAD(q))
                g.set_edge_type(e, EdgeType.SIMPLE)
            elif (g.type(v) == VertexType.Z or g.type(v) == VertexType.X) and g.vertex_degree(v) == 2:
                ns = list(g.neighbors(v))
                w = ns[0] if ns[1] == o else ns[1]
                progress = True

                if g.phase(v) != 0:
                    gate = (ZPhase(q, g.phase(v)) if g.type(v) == VertexType.Z else
                            XPhase(q, g.phase(v)))
                    circ.prepend_gate(gate)

                g.add_edge((w,o), edgetype=g.edge_type(g.edge(w,v)))
                g.remove_vertex(v)
                
        if progress: continue
        
        for q1,o1 in enumerate(outputs):
            for q2,o2 in enumerate(outputs):
                if o1 == o2: continue
                v1 = list(g.neighbors(o1))[0]
                v2 = list(g.neighbors(o2))[0]
                if g.connected(v1,v2):
                    if ((g.type(v1) == g.type(v2) and g.edge_type(g.edge(v1,v2)) == EdgeType.SIMPLE) or
                        (g.type(v1) != g.type(v2) and g.edge_type(g.edge(v1,v2)) == EdgeType.HADAMARD)):
                        raise ValueError("ZX diagram is not unitary")
                    
                    if g.type(v1) == VertexType.Z and g.type(v2) == VertexType.X:
                        # CNOT
                        progress = True
                        circ.prepend_gate(CNOT(control=q1,target=q2))
                        g.remove_edge(g.edge(v1,v2))
                    elif g.type(v1) == VertexType.Z and g.type(v2) == VertexType.Z:
                        # CZ
                        progress = True
                        circ.prepend_gate(CZ(control=q1,target=q2))
                        g.remove_edge(g.edge(v1,v2))
                    elif g.type(v1) == VertexType.X and g.type(v2) == VertexType.X:
                        # conjugate CZ
                        progress = True
                        circ.prepend_gate(XCX(control=q1, target=q2))
                        g.remove_edge(g.edge(v1,v2))

    return graph_to_swaps(g, up_to_perm) + circ


def graph_to_swaps(g: BaseGraph[VT, ET], no_swaps: bool = False) -> Circuit:
    """This function expects a graph like circuit as input. Converts a graph containing only normal and Hadamard edges (i.e., no vertices other than
    inputs and outputs) into a circuit of Hadamard and SWAP gates. If 'no_swaps' is True, only add
    Hadamards where needed"""
    swap_map = {}
    leftover_swaps = False
    inputs = g.inputs()
    outputs = g.outputs()

    if not is_graph_like(g):
        raise ValueError("Input graph is not graph-like")

    c = Circuit(len(inputs))

    for q,v in enumerate(outputs): # check for a last layer of Hadamards, and see if swap gates need to be applied.
        inp = list(g.neighbors(v))[0]
        if inp not in inputs: 
            raise TypeError("Algorithm failed: Graph is not fully reduced")
            return c
        if g.edge_type(g.edge(v,inp)) == EdgeType.HADAMARD:
            c.prepend_gate(HAD(q))
            g.set_edge_type(g.edge(v,inp),EdgeType.SIMPLE)
        q2 = inputs.index(inp)
        if q2 != q: leftover_swaps = True
        swap_map[q] = q2
    if not no_swaps and leftover_swaps:
        for t1, t2 in permutation_as_swaps(swap_map):
            c.prepend_gate(SWAP(t1, t2))
    return c


def extract_clifford_normal_form(g: BaseGraph[VT,ET]) -> Circuit:
    """Given a Clifford graph, extracts a circuit that follows the normal form described in
    *Graph-theoretic Simplification of Quantum Circuits with the ZX-calculus* (https://arxiv.org/abs/1902.03178).
    That is, a circuit consisting of layers Had-Phase-CZ-CNOT-Had-CZ-Phase-Had.
    """
    # We prepare the circuit in the same way as we do in simplify.to_clifford_normal_form_graph()
    # To make it foolproof, we process it in the desired way first.
    full_reduce(g)
    g.normalize()
    # At this point the only vertices g should have are those directly connected to an input or an output (and not both).
    if any([((g.phase(v)*4) % 2 != 0) for v in g.vertices()]):  # If any phase is not a multiple of 1/2, then this will fail.
        raise ValueError("Specified graph is not Clifford.")
    
    inputs = list(g.inputs())
    outputs = list(g.outputs())
    v_inputs = [list(g.neighbors(i))[0] for i in inputs] # input vertices should have a unique spider neighbor
    v_outputs = [list(g.neighbors(o))[0] for o in outputs] # input vertices should have a unique spider neighbor

    if len(inputs) != len(outputs):
        raise ValueError("Number of input wires does not match number of output wires. Currently only unitary Clifford extraction is supported.")
    if len(v_inputs) != len(inputs) or len(v_inputs) != len(v_outputs):
        raise ValueError("Something has gone wrong with simplifying the Clifford diagram to the graph normal form.")
    
    c = Circuit(len(inputs))

    for q in range(len(inputs)):
        if g.edge_type(g.edge(inputs[q],v_inputs[q])) == EdgeType.HADAMARD:
            c.add_gate(HAD(q))

    for q in range(len(inputs)):
        phase = g.phase(v_inputs[q])
        if phase != 0:
            c.add_gate(ZPhase(q,phase))

    for q1 in range(len(inputs)):
        for q2 in range(q1+1, len(inputs)):
            if g.connected(v_inputs[q1],v_inputs[q2]):
                c.add_gate(CZ(q1,q2))

    adj = bi_adj(g, v_outputs, v_inputs)
    for cnot in adj.to_cnots(use_log_blocksize=True):
        c.add_gate(cnot)

    for q in range(len(outputs)):
        c.add_gate(HAD(q))

    for q1 in range(len(outputs)):
        for q2 in range(q1+1, len(outputs)):
            if g.connected(v_outputs[q1],v_outputs[q2]):
                c.add_gate(CZ(q1,q2))

    for q in range(len(outputs)):
        phase = g.phase(v_outputs[q])
        if phase != 0:
            c.add_gate(ZPhase(q,phase))

    for q in range(len(outputs)):
        if g.edge_type(g.edge(outputs[q],v_outputs[q])) == EdgeType.HADAMARD:
            c.add_gate(HAD(q))

    return c


class LookaheadNode:
    """
    A class for the lookahead extraction.

    Performs most operations recursively, such as expanding nodes,
    picking next roots, and finding optimal nodes.

    Supports two structures:
        - for CNOT optimisation, each node holds the part of the circuit that it creates
        - for depth optimisation, only the leaves hold circuits

    The correct order of operations for a root is:
    root.expand -> root.get_finished -> root.next_nodes, as used in lookahead_extract_base
    """

    def __init__(self,
                 g: BaseGraph[VT, ET],
                 c: Circuit,
                 frontier: List[VT],
                 qubit_map: Dict[VT, int],
                 gadgets: Dict[VT, VT],
                 opt_depth: bool,
                 hard_limit: int,
                 ext_count: int = 0):
        self.g: BaseGraph[VT, ET] = g
        self.c: Circuit = c
        self.frontier: List[VT] = frontier
        self.qubit_map: Dict[VT, int] = qubit_map
        self.gadgets: Dict[VT, VT] = gadgets
        self.children: List['LookaheadNode'] = []
        self.expanded: bool = False
        self.ext_count: int = ext_count
        self.collected: bool = False
        self.opt_depth: bool = opt_depth
        self.d: int = -1
        self.total_d: int = -1
        self.finished_children: Optional[List[int]] = None
        self.hard_limit: int = hard_limit

    def update_hard_limit(self, new_limit: int):
        self.hard_limit = new_limit
        for child in self.children:
            child.update_hard_limit(new_limit)

    def mark_expanded(self):
        """
        Free the memory once the node is expanded.
        """
        self.expanded = True
        self.g = None
        if self.opt_depth:
            self.c = None
        self.frontier = None
        self.qubit_map = None
        self.gadgets = None

    def stats(self):
        """
        Returns the number of nodes, leaves, and maximum depth
        """
        if len(self.children) == 0:
            return 1, 1, 0
        depth = 0
        nodes = 1
        leaves = 0
        for child in self.children:
            c_n, c_l, c_d = child.stats()
            if c_d > depth:
                depth = c_d
            nodes += c_n
            leaves += c_l
        return nodes, leaves, depth + 1

    def optimal(self, rp: 'RootPicker', d: int = -1):
        """
        Computes the stats for leaves (number of 2 qubit gates or depth) to determine the best nodes.

        Args:
            rp: RootPicker, used to collect the leaves and filter the best ones
            d: for CNOT optimisation only, the total number of two qubit gates of the parent nodes, up to root
        """
        if self.collected:
            # Already checked, not interested in it
            return
        if not self.opt_depth:
            if self.d == -1:
                self.d = get_optimize_value(self.c, False)
            if d == -1:
                d = 0
            d = d + self.d
            self.total_d = d
        if len(self.children) == 0:
            if self.opt_depth:
                if self.total_d == -1:
                    self.total_d = get_optimize_value(self.c, True)
                d = self.total_d
            if -1 < self.hard_limit <= d:
                return
            rp.add_leaf(self, Fraction(d, self.ext_count))
            return
        for child in self.children:
            child.optimal(rp, d)

    def next_nodes(self, min_extracted: int, rp: 'RootPicker', prev_circ: Optional[Circuit] = None, d: int = -1):
        """
        Used to find the next roots. Looks for nodes that fit the parameters, then searches their leaves.

        Args:
            min_extracted: minimum number of vertices extracted for a node to be considered a root
            rp: RootPicker to collect the possible roots and best leaves
            prev_circ: for CNOT optimisation, collect the circuit starting from the root up to the node to assign
            to the node in case it is selected as a next root
            d: for CNOT optimisation, the total number of two qubit gates starting from the root
        """
        if len(self.children) == 0 or self.ext_count > min_extracted:
            rp.add_possible_root(self, prev_circ, d)
            self.optimal(rp, d)
            return
        if not self.opt_depth:
            if self.d == -1:
                self.d = get_optimize_value(self.c, False)
            prev_circ = self.c if prev_circ is None else prev_circ + self.c
            d = self.d if d == -1 else d + self.d
            self.total_d = d
        for child in self.children:
            child.next_nodes(min_extracted, rp, prev_circ, d)

    def __has_finished(self) -> bool:
        """
        For CNOT optimisation, check which children of a node has finished leaves
        """
        if self.collected:
            return False
        if not self.expanded:
            if len(self.frontier) == 0:
                return True
            return False
        for i, child in enumerate(self.children):
            fin = child.__has_finished()
            if fin:
                if self.finished_children is None:
                    self.finished_children = []
                self.finished_children.append(i)
        return self.finished_children is not None

    def __collect_finished_cnot(self, best_c: Optional[Circuit], best_d: int,
                                c: Circuit, up_to_perm: bool) -> Tuple[Optional[Circuit], int]:
        """
        Find the best fully extracted circuits in the CNOT optimisation case
        """
        c = c + self.c
        if not self.expanded:
            self.collected = True
            id_simp(self.g, quiet=True)
            c.gates = list(reversed(c.gates))
            c = graph_to_swaps(self.g, up_to_perm) + c
            d = get_optimize_value(c, self.opt_depth, True)
            if d < best_d or best_d == -1:
                best_c = c
                best_d = d
        else:
            if self.finished_children is None:
                raise AssertionError("LookaheadNode.__collect_finished_cnot was called before " +
                                     "LookaheadNode.__has_finished")
            for i in self.finished_children:
                best_c, best_d = self.children[i].__collect_finished_cnot(best_c, best_d, c, up_to_perm)
            self.finished_children = None
        return best_c, best_d

    def __collect_finished_depth(self, best_c: Optional[Circuit], best_d: int,
                                 up_to_perm: bool) -> Tuple[Optional[Circuit], int]:
        """
        Find the best fully extracted circuits in the depth optimisation case
        """
        if not self.expanded and not self.collected and len(self.frontier) == 0:
            self.collected = True
            id_simp(self.g, quiet=True)
            self.c.gates = list(reversed(self.c.gates))
            self.c = graph_to_swaps(self.g, up_to_perm) + self.c
            d = get_optimize_value(self.c, self.opt_depth, True)
            if d < best_d or best_d == -1:
                best_c = self.c
                best_d = d
        elif self.children is not None:
            for child in self.children:
                best_c, best_d = child.__collect_finished_depth(best_c, best_d, up_to_perm)
        return best_c, best_d

    def get_finished(self, best_c: Optional[Circuit], best_d: int, up_to_perm: bool) -> Tuple[Optional[Circuit], int]:
        """
        Find the best fully extracted circuits in the CNOT optimisation case

        Args:
            best_c: best circuit up to this point
            best_d: best two qubit count/depth up to this point
            up_to_perm: return an equivalent circuit up to a permutation of inputs

        Returns:
            best_c, best_d updated with any finished leaves found
        """
        if self.opt_depth:
            return self.__collect_finished_depth(best_c, best_d, up_to_perm)
        if self.__has_finished():
            return self.__collect_finished_cnot(best_c, best_d, Circuit(self.c.qubits), up_to_perm)
        return best_c, best_d

    def can_expand(self):
        if not self.expanded:
            if len(self.frontier) == 0:
                return False
            if self.d == -1:  # Should only be the case for the initial root
                return True
            if self.total_d >= self.hard_limit > -1:
                return False
            return True
        for child in self.children:
            if child.can_expand():
                return True
        return False

    def branch_child(self) -> 'LookaheadNode':
        child = LookaheadNode(self.g.clone(), self.c.copy() if self.opt_depth else Circuit(self.c.qubits),
                              self.frontier.copy(), self.qubit_map.copy(), self.gadgets.copy(), self.opt_depth,
                              self.hard_limit, self.ext_count)
        self.children.append(child)
        return child

    def apply_cnots(self, cnots: List[CNOT], m: Mat2, neighbors: List[VT]):
        self.ext_count += apply_cnots(self.g, self.c, self.frontier, self.qubit_map, cnots, m, neighbors)

    def expand(self, limit: int, max_depth: int, algorithms: List[int]):
        if max_depth == 0:
            return
        if self.total_d >= self.hard_limit > -1:
            return
        while not self.expanded and self.ext_count < limit and len(self.frontier) != 0:
            self.d = -1
            clean_frontier(self.g, self.c, self.frontier, self.qubit_map)

            # Now we can proceed with the actual extraction
            # First make sure that frontier is connected in correct way to inputs
            neighbor_set = neighbors_of_frontier(self.g, self.frontier)

            if not self.frontier:
                break  # No more vertices to be processed. We are done.

            # First we check if there is a phase gadget in the way
            if remove_gadget(self.g, self.frontier, self.qubit_map, neighbor_set, self.gadgets):
                # There was a gadget in the way. Go back to the top
                continue

            neighbors = list(neighbor_set)
            m = bi_adj(self.g, neighbors, self.frontier)

            cnots: List[CNOT] = []
            if all(sum(row) != 1 for row in m.data):  # No easy vertex

                qubits = self.c.qubits
                branches: List[List[CNOT]] = []
                for alg in algorithms:
                    # Try different algorithms to get distinct sets of CNOTs that can be used
                    cnots_opt = self.apply_operation(alg, m, neighbors)
                    if cnots_opt is None:
                        continue
                    should_append = True
                    for i in range(len(branches)):
                        if compare_cnots(cnots_opt, branches[i], qubits):  # Same result when applying CNOTs
                            should_append = False
                            if len(cnots_opt) < len(branches[i]):
                                branches[i] = cnots_opt
                            break
                    if should_append:
                        branches.append(cnots_opt)

                if len(branches) == 0:
                    raise Exception("All steps returned impossible")
                elif len(branches) == 1:  # No need to branch
                    cnots = branches[0]
                else:  # Branch and go to children
                    for cnots_opt in branches:
                        child = self.branch_child()
                        child.apply_cnots(cnots_opt, m.copy(), neighbors)
                    self.mark_expanded()
                    break

            self.apply_cnots(cnots, m, neighbors)

        for child in self.children:
            child.expand(limit, max_depth - 1, algorithms)

    def apply_operation(self, operation_id: int, m: Mat2, neighbors: List[VT]) -> Optional[List[CNOT]]:
        """
        Apply one of the possible operations to the current node to obtain a list of CNOTs
        """
        cnots: List[CNOT]

        if operation_id == 0:
            perm = column_optimal_swap(m)
            perm = {v: k for k, v in perm.items()}
            neighbors2 = [neighbors[perm[i]] for i in range(len(neighbors))]
            m2 = bi_adj(self.g, neighbors2, self.frontier)
            cnots = m2.to_cnots(optimize=False, use_log_blocksize=True)
            cnots = filter_duplicate_cnots(
                cnots)  # Since the matrix is not square, the algorithm sometimes introduces duplicates

        elif operation_id == 1:
            greedy_operations = greedy_reduction(m)
            if greedy_operations is None:
                return None
            cnots = [CNOT(target, control) for control, target in greedy_operations]

        elif operation_id == 2:
            greedy_operations = greedy_reduction_flat(m)
            if greedy_operations is None:
                return None
            cnots = [CNOT(target, control) for control, target in greedy_operations]

        elif operation_id == 3:
            greedy_operations = greedy_two_reduction(m)
            if greedy_operations is None:
                return None
            cnots = [CNOT(target, control) for control, target in greedy_operations]

        else:
            raise Exception("Unknown extraction step: {}".format(operation_id))

        cnots_opt = remove_extra_cnots(cnots, m)
        return cnots_opt


class RootPicker:
    """
    Class for picking the next roots that correspond to the best k leaves.
    """
    def __init__(self, k: int):
        self.k: int = k
        self.nodes: List[Optional[Tuple[LookaheadNode, Optional[Circuit], int]]] = []
        self.best: List[Tuple[LookaheadNode, Fraction, int]] = []

    def add_possible_root(self, n: LookaheadNode, c: Optional[Circuit], d: int):
        """
        Add a node that satisfies the conditions for being a root in the next step
        For CNOT optimisation, include the circuit up to this node
        """
        # First, remove any of the previous nodes that we do not need
        # This will let the garbage collector to clean the unnecessary nodes
        if len(self.nodes) > 0:
            s = set()
            for p in self.best:
                s.add(p[2])
            for i in range(len(self.nodes)):
                if i not in s:
                    self.nodes[i] = None
        # Add the new node
        self.nodes.append((n, c, d))

    def add_leaf(self, n: LookaheadNode, d: Fraction):
        """
        Add a leaf to the list of the best k leaves. It is assigned to the last possible root added.

        Args:
            n: the leaf
            d: the metric for comparison (two qubit count or depth divided by the number of vertices extracted)
        """
        if len(self.nodes) == 0:
            raise AssertionError("Adding leaf without any possible root")
        i = 0
        while i < len(self.best) and d >= self.best[i][1]:
            i += 1
        if i == self.k:
            return
        self.best.insert(i, (n, d, len(self.nodes) - 1))
        if len(self.best) > self.k:
            self.best.pop()

    def get_next_roots(self) -> List[Optional[LookaheadNode]]:
        """
        Pick the roots that correspond to the best k leaves added

        The type needs Optional to match the type in `lookahead_extract_base`, which allows nodes to be removed from
        the list early.
        """
        if len(self.nodes) == 0:
            return []
        s = set()
        for p in self.best:
            s.add(p[2])
        nodes: List[Optional[LookaheadNode]] = []
        for i in s:
            entry = self.nodes[i]
            if entry is None:
                # Does not happen, needed for type checking
                raise AssertionError("RootPicker removed a root that was still needed")
            else:
                n = entry[0]
                if entry[1] is not None:
                    n.c = entry[1] + n.c
                    if n.d != -1:
                        n.d = entry[2] + n.d
                nodes.append(n)
        return nodes


def lookahead_extract_base(
        g: BaseGraph[VT, ET],
        steps: int = -1,  # ideal number of steps to look ahead
        depth_limit: int = 7,  # maximum depth os the lookahead tree
        min_extract: int = 5,  # minimum extracted vertices at each step, improves performance
        nodes_kept: int = 4,   # keep the best 'nodes_kept' nodes when advancing to the next step
        hard_limit: int = -1,  # do not consider circuits where the comparison metric is over this
        algorithms: Optional[List[int]] = None,  # always include 0, pick any from 1, 2, 3
        optimize_for_depth: bool = False,  # optimize for depth instead of two qubit gates
        compare_basic: bool = True,  # use the default extractions and pick the best
        up_to_perm: bool = False  # return an equivalent circuit up to an input permutation
        ) -> Optional[Circuit]:
    """
    Main method for the lookahead extraction. Uses different methods to produce CNOTS and extract vertices,
    simulates a few steps in advance, and picks the best results.

    Args:
        g: the graph to transform into a circuit
        steps: the number of vertices to extract before comparing different results; should generally vary with the number of qubits in the graph
        depth_limit: the maximum depth of the search tree, to stop from memory problems in some edge cases; negative values remove the constrint
        min_extract: the minimum number of vertices extracted for a node to be considered as a root for the next step
        nodes_kept: instead of only considering the best result when looking ahead, consider the best 'nodes_kept' results and pick the nodes that correspond to each result to be roots in the next step
        hard_limit: stop the search when reaching 'hard_limit' two qubit gates/depth
        algorithms: the different algorithms to use in the search; always include 0, other possibilities are 1, 2, 3; the length of this list gives the branching factor of the search tree
        optimize_for_depth: if set to false (default), optimize for the number of two qubit gates; if set to true, optimize for depth
        compare_basic: perform the standard extractions and pick the best between the standard and the result of the lookahead extraction
        up_to_perm: if set to true, returns a circuit that corresponds to the graph up to a permutation of th inputs

    Returns:
        A circuit that corresponds to the given graph, with two qubit count / depth less than 'hard_limit'; None if no such circuit was found
    """

    if steps < 1:
        steps = len(g.inputs()) * 3
    if depth_limit < 1:
        depth_limit = -1
    if min_extract < 0:
        min_extract = 0
    if nodes_kept < 1:
        nodes_kept = 1
    if hard_limit < 0:
        hard_limit = -1
    if algorithms is None:
        algorithms = [0, 1, 2]

    best_c = None
    best_d = hard_limit

    if compare_basic:
        # Perform the basic extractions and pick the best result
        c1 = extract_circuit(g.clone(), optimize_cnots=1, up_to_perm=up_to_perm)
        d1 = get_optimize_value(c1, optimize_for_depth, True)
        if best_d > d1 or best_d == -1:
            best_c = c1
            best_d = d1
        c1 = extract_circuit(g.clone(), optimize_cnots=3, up_to_perm=up_to_perm)
        d1 = get_optimize_value(c1, optimize_for_depth, True)
        if best_d > d1 or best_d == -1:
            best_c = c1
            best_d = d1
        hard_limit = best_d

    gadgets = {}
    inputs = g.inputs()
    outputs = g.outputs()
    for v in g.vertices():
        if g.vertex_degree(v) == 1 and v not in inputs and v not in outputs:
            n = list(g.neighbors(v))[0]
            gadgets[n] = v

    qubit_map: Dict[VT, int] = dict()
    frontier = []
    for i, o in enumerate(outputs):
        v = list(g.neighbors(o))[0]
        if v in inputs:
            continue
        frontier.append(v)
        qubit_map[v] = i

    roots: List[Optional[LookaheadNode]] =\
        [LookaheadNode(g, Circuit(len(inputs)), frontier, qubit_map, gadgets, optimize_for_depth, hard_limit)]

    while len(roots) > 0:
        rp = RootPicker(nodes_kept)
        for i in range(len(roots)):
            root = roots[i]
            if root is None:
                continue  # Never happens, but creates problems with type checker
            if root.hard_limit > hard_limit:
                root.update_hard_limit(hard_limit)
            if root.can_expand():
                new_limit = root.ext_count + steps
                prev_extracted = root.ext_count
                root.expand(new_limit, depth_limit, algorithms)
                best_c, best_d = root.get_finished(best_c, best_d, up_to_perm)
                if best_d < hard_limit:
                    hard_limit = best_d
                    root.update_hard_limit(hard_limit)
                root.next_nodes(prev_extracted + min_extract, rp)
            # Allow unneeded nodes to be removed to free memory
            roots[i] = None
        roots = rp.get_next_roots()

    return best_c


def get_optimize_value(c: Circuit, optimize_for_depth: bool, expand_to_basic: bool = False) -> int:
    """ Computes the two qubit count or the depth for the circuit. """
    if expand_to_basic:
        c = c.to_basic_gates()
    if not optimize_for_depth:
        d = 0
        for gate in c.gates:
            if isinstance(gate, (CNOT, CZ)):
                d += 1
        return d
    return c.depth()


def cnots_to_xor_list(cnots: List[CNOT], size: int) -> List[Set[int]]:
    sets = [{i} for i in range(size)]
    for c in cnots:
        for i in sets[c.target]:
            if i in sets[c.control]:
                sets[c.control].remove(i)
            else:
                sets[c.control].add(i)
    return sets


def compare_cnots(c1: List[CNOT], c2: List[CNOT], size: int) -> bool:
    """ Checks if two sets of cnots give the same results """
    s1 = cnots_to_xor_list(c1, size)
    s2 = cnots_to_xor_list(c2, size)
    for index in range(size):
        if s1[index] != s2[index]:
            return False
    return True


def lookahead_fast(g: BaseGraph[VT, ET], optimize_for_depth: bool = False, up_to_perm: bool = False) -> Circuit:
    """
    A lookahead extraction with relatively fast results. For details see :func:`lookahead_extract_base`
    """
    c = lookahead_extract_base(g, 4 * len(g.inputs()), 8, 5, 4, -1, [0, 1], optimize_for_depth, False, up_to_perm)
    if c is None:
        raise AssertionError("Lookahead extraction with no hard limit returned None")
    return c


def lookahead_extract(g: BaseGraph[VT, ET], optimize_for_depth: bool = False, up_to_perm: bool = False) -> Circuit:
    """
        A lookahead extraction with recommended parameters. For details see :func:`lookahead_extract_base`
    """
    qubits = len(g.inputs())
    c = lookahead_extract_base(g.clone(), 4 * qubits, 8, 0, 4, -1, [0, 1], optimize_for_depth, True, up_to_perm)
    if c is None:
        raise AssertionError("Lookahead extraction with no hard limit returned None")
    d = get_optimize_value(c, optimize_for_depth, True)
    c1 = lookahead_extract_base(g, 4 * qubits, 8, 5, 4, d, [0, 3], optimize_for_depth, False, up_to_perm)
    if c1 is not None:
        d1 = get_optimize_value(c1, optimize_for_depth, True)
        if d1 < d:
            c = c1
    return c


def lookahead_full(g: BaseGraph[VT, ET], optimize_for_depth: bool = False, up_to_perm: bool = False) -> Circuit:
    """
        A lookahead extraction which compares a number of possible extractions and returns the best result.
        Can take a very long time for large circuits. For details see :func:`lookahead_extract_base`
    """
    qubits = len(g.inputs())
    c = lookahead_extract_base(g.clone(), 3 * qubits, 7, qubits, 4, -1,
                               [0, 1, 3], optimize_for_depth, True, up_to_perm)
    if c is None:
        raise AssertionError("Lookahead extraction with no hard limit returned None")
    d = get_optimize_value(c, optimize_for_depth, True)
    c1 = lookahead_extract_base(g.clone(), 4 * qubits, 8, 0, 4, d, [0, 1], optimize_for_depth, False, up_to_perm)
    if c1 is not None:
        d1 = get_optimize_value(c1, optimize_for_depth, True)
        if d1 < d:
            c = c1
            d = d1
    c1 = lookahead_extract_base(g.clone(), 4 * qubits, 8, 5, 4, d, [0, 2], optimize_for_depth, False, up_to_perm)
    if c1 is not None:
        d1 = get_optimize_value(c1, optimize_for_depth, True)
        if d1 < d:
            c = c1
            d = d1
    c1 = lookahead_extract_base(g, 4 * qubits, 8, 5, 4, d, [0, 3], optimize_for_depth, False, up_to_perm)
    if c1 is not None:
        d1 = get_optimize_value(c1, optimize_for_depth, True)
        if d1 < d:
            c = c1
    return c



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List

from .simplify import simp, Stats, spider_simp, id_simp
from .simplify import pivot_simp, lcomp_simp, pivot_gadget_simp, pivot_boundary_simp
from .utils import EdgeType, VertexType
from .graph.base import BaseGraph, VT,ET
from .rewrite_rules.hrules import *
from .rewrite_rules.rules import MatchObject


def hadamard_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=False) -> int:
    """Converts as many Hadamards represented by H-boxes to Hadamard-edges."""
    # We can't use the regular simp function, because removing H-nodes could lead to an infinite loop,
    # since sometimes g.add_edge_table() decides that we can't change an H-box into an H-edge.
    i = 0
    while True:
        vcount = g.num_vertices()
        m = match_hadamards(g, matchf)
        if len(m) == 0: break
        i += 1
        if i == 1 and not quiet: print("hadamard_simp: ",end='')
        if not quiet: print(len(m), end='')
        etab, rem_verts, rem_edges, check_isolated_vertices = hadamard_to_h_edge(g, m)
        g.add_edge_table(etab)
        g.remove_edges(rem_edges)
        g.remove_vertices(rem_verts)
        if check_isolated_vertices: g.remove_isolated_vertices()
        if not quiet: print('. ', end='')
        if g.num_vertices() >= vcount: break # To make sure we don't get in an infinite loop
    if not quiet and i>0: print(' {!s} iterations'.format(i))
    return i

def to_hypergraph_form(g: BaseGraph[VT,ET]) -> None:
    """Convert a Graph g to hypergraph form (see https://arxiv.org/abs/2003.13564)
    First, all X spiders are turned into Z-spiders by color-change, 
    then all interior Hadamard edges are replaced by arity-2 hboxes 
    and all Z-phases are replaced by arity-1 hboxes.
    Finally, to ensure H-boxes are only connected to white spiders, 
    additional identities are introduced"""
    to_gh(g)
    del_e: List[ET] = []
    add_e: List[Tuple[VT,VT]] = []
    types = g.types()
    phases = g.phases()
    es = list(g.edges())
    vs = list(g.vertices())
    for v in vs:
        if types[v] == VertexType.Z and phases[v] != 0:
            h = g.add_vertex(VertexType.H_BOX)
            g.add_edge((h,v))
            g.set_qubit(h, g.qubit(v) - 0.5)
            g.set_row(h, g.row(v) + 0.5)
            g.set_phase(h, phases[v])
            g.set_phase(v, 0)
    for e in es:
        if g.edge_type(e) == EdgeType.HADAMARD:
            s,t = g.edge_st(e)
            rs = g.row(s)
            rt = g.row(t)
            qs = g.qubit(s)
            qt = g.qubit(t)
            if types[s] == VertexType.BOUNDARY or types[t] == VertexType.BOUNDARY: continue
            h = g.add_vertex(VertexType.H_BOX)
            del_e.append(e)
            add_e.append((s, h))
            add_e.append((h, t))
            g.scalar.add_power(-1) # Correct for sqrt(2) scalar difference in H-edge and H-box.
            if qs == qt:
                g.set_qubit(h, qs)
            else:
                q = (qs + qt) / 2
                if round(q) == q: q += 0.5
                g.set_qubit(h, q)
            g.set_row(h, (rs + rt) / 2)
    g.remove_edges(del_e)
    g.add_edges(add_e)
    # At this point, all interior Hadamard edges and phases are converted to H-boxes
    # It remains to add missing identities between connected H-boxes
    for v in list(g.vertices()):
        if types[v] != VertexType.H_BOX: continue
        del_e = []
        add_e = []
        for n in g.neighbors(v):
            if types[n] != VertexType.H_BOX: continue
            w = g.add_vertex(VertexType.Z)
            g.set_row(w, (g.row(v)+g.row(n))/2)
            g.set_qubit(w,(g.qubit(v)+g.qubit(n))/2)
            del_e.append(g.edge(v,n))
            add_e.append((v,w))
            add_e.append((w,n))
        g.remove_edges(del_e)
        g.add_edges(add_e)

def from_hypergraph_form(g: BaseGraph[VT,ET]) -> None:
    """Convert a graph with hboxes and no interior Hadamard edges back to 'ZX-friendly'
    form. All arity-2 hboxes with phase pi are turned into Hadamard edges and all
    arity-1 hboxes connected to Z-spiders are fused in. Note that more general hboxes
    are not changed, so this does *not* give a ZX graph in general."""
    types = g.types()
    phases = g.phases()
    hs = [h for h in g.vertices() if types[h] == VertexType.H_BOX]
    for h in hs:
        if g.vertex_degree(h) == 1:
            n = next(iter(g.neighbors(h)))
            if types[n] == VertexType.Z:
                g.set_phase(n, phases[n] + phases[h])
                g.remove_vertex(h)
    hadamard_simp(g,quiet=True)


def par_hbox_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=False, stats:Optional[Stats]=None) -> int:
    return simp(g, 'par_hbox_simp', match_par_hbox, par_hbox, matchf=matchf, quiet=quiet, stats=stats)

def par_hbox_intro_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=False, stats:Optional[Stats]=None) -> int:
    return simp(g, 'par_hbox_intro_simp', match_par_hbox_intro, par_hbox_intro, matchf=matchf, quiet=quiet, stats=stats)

def copy_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=False, stats:Optional[Stats]=None) -> int:
    return simp(g, 'copy_simp', match_copy, apply_copy, matchf=matchf, quiet=quiet, stats=stats)

def hspider_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[ET],bool]]=None, quiet:bool=False, stats:Optional[Stats]=None) -> int:
    return simp(g, 'hspider_simp', match_connected_hboxes, fuse_hboxes, matchf=matchf, quiet=quiet, stats=stats)

def hbox_parallel_not_remove_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=False, stats:Optional[Stats]=None) -> int:
    return simp(g, 'hbox_parallel_not_remove_simp', match_hbox_parallel_not, hbox_parallel_not_remove, matchf=matchf, quiet=quiet, stats=stats)


# a stripped-down version of "simp", since hrules don't return edge tables etc
def hsimp(
        g: BaseGraph[VT,ET], 
        name:str, 
        match: Callable[..., List[MatchObject]],
        rule: Callable[[BaseGraph[VT,ET],List[MatchObject]],None], 
        iterations:int=-1, 
        quiet:bool=False
        ) -> int:
    i = 0
    while iterations == -1 or i < iterations:
        ms = match(g)
        if len(ms) > 0:
            rule(g, ms)
            i += 1
            if i == 1 and not quiet: print("{}: ".format(name),end='')
            if not quiet: print(len(ms), end='. ')
        else:
            break
    if not quiet and i>0: print(' {!s} iterations'.format(i))
    return i

def hpivot_simp(g: BaseGraph[VT,ET], quiet:bool=False) -> int:
    spider_simp(g, quiet=quiet)
    id_simp(g, quiet=quiet)
    from_hypergraph_form(g)
    spider_simp(g, quiet=quiet)
    id_simp(g, quiet=quiet)
    count = 0
    while True:
        to_hypergraph_form(g)
        i = hsimp(g, 'hpivot', match_hpivot, hpivot, iterations=1, quiet=quiet)
        #i += hsimp(g, 'zero_hbox', match_zero_hbox, zero_hbox, quiet=quiet)
        i += par_hbox_simp(g,quiet=quiet)
        from_hypergraph_form(g)
        spider_simp(g, quiet=quiet)
        id_simp(g, quiet=quiet)
        if i == 0: break
        count += 1
    return count

def zh_simp(g: BaseGraph[VT,ET], quiet:bool=False) -> int:
    """Does a whole bunch of rewrites to simplify diagrams containing
    arbitrary Z-spiders, X-spiders and H-boxes. 
    Tries to do as many "non-complicating" rewrites before doing
    rewrites that can make a diagram look more complex."""
    count = 0
    while True:
        to_gh(g)
        i = spider_simp(g, quiet=quiet)
        i += id_simp(g, quiet=quiet)
        from_hypergraph_form(g)
        i += spider_simp(g, quiet=quiet)
        i += id_simp(g, quiet=quiet)
        i += hspider_simp(g, quiet=quiet)
        if i > 0: 
            count += 1
            continue
        if copy_simp(g, quiet=quiet):
            count += 1
            continue
        if par_hbox_simp(g,quiet=quiet):
            count += 1
            continue
        if par_hbox_intro_simp(g,quiet=quiet):
            count += 1
            continue
        if hbox_parallel_not_remove_simp(g,quiet=quiet):
            count += 1
            continue
        i = pivot_simp(g, quiet=quiet)
        i += lcomp_simp(g, quiet=quiet)
        if i> 0: 
            count += 1
            continue
        if pivot_gadget_simp(g, quiet=quiet):
            count += 1
            continue
        if pivot_boundary_simp(g, quiet=quiet):
            count += 1
            continue

        to_hypergraph_form(g)
        i = hsimp(g, 'hpivot', match_hpivot, hpivot, iterations=1, quiet=quiet)
        i += par_hbox_simp(g,quiet=quiet)
        from_hypergraph_form(g)
        id_simp(g, quiet=True)
        if i > 0:
            count += 1
            continue
        
        break
    return count



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import sys
import numpy as np
if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')

from pyzx_param.linalg import Mat2, rank_factorize, generalized_inverse


class TestMat2(unittest.TestCase):

    def setUp(self):
        self.m1 = Mat2([[1,0],
                        [1,1]])
        self.m2 = Mat2([[1,1],
                        [1,1]])
        self.m3 = Mat2([[1,0,1,1,0],
                        [1,1,1,0,0],
                        [1,1,0,0,1],
                        [0,1,0,1,0],
                        [0,0,1,1,0]])
        self.m4 = Mat2([[1,0,1,0,0],
                        [0,1,1,0,0],
                        [1,1,0,0,1],
                        [0,1,0,1,0],
                        [0,0,0,1,1]])

    def test_matrix_multiplication(self):
        result = Mat2([[1,1],[0,0]])
        self.assertEqual(self.m1*self.m2, result)
        result = Mat2([[0,1],[0,1]])
        self.assertEqual(self.m2*self.m1, result)

    def test_gauss_makes_upper_triangular(self):
        self.m3.gauss()
        flagged = False
        for i in range(self.m3.rows()):
            for j in range(min(self.m3.cols(),i)):
                if self.m3.data[i][j] != 0: flagged = True
        self.assertFalse(flagged)

    def test_rank_of_matrix(self):
        self.assertEqual(self.m3.rank(),4)

    def test_non_full_rank_matrix_shouldnt_have_inverse(self):
        self.assertIsNone(self.m3.inverse())

    def test_inverse(self):
        inv = self.m4.inverse()
        self.assertEqual(inv*self.m4, Mat2.id(5))
        self.assertEqual(self.m4*inv, Mat2.id(5))

    def test_solve_inhomogeneous_equation(self):
        b = Mat2([[1],[1],[0],[0],[0]])
        x = self.m3.solve(b)
        self.assertEqual(self.m3*x, b)
        b = Mat2([[1],[0],[1],[1],[0]])
        x = self.m4.solve(b)
        self.assertEqual(self.m4*x, b)

    def test_factor(self):
        m0, m1 = self.m3.factor()
        self.assertEqual(m0.cols(),self.m3.rank())
        self.assertEqual(m1.rows(),self.m3.rank())
        self.assertEqual(m0*m1, self.m3)

    def test_rank_factorize(self):
        A = np.array([[1, 1, 1, 1, 1],
                      [1, 1, 1, 1, 1],
                      [0, 1, 0, 0, 1],
                      [1, 0, 1, 1, 0]], dtype=np.int8)
        r, U, V = rank_factorize(A)
        assert r == 2
        assert ((U[:, :r] @ V[:r]) % 2 == A).all()

    def test_generalized_inverse(self):
        A = np.array([[1, 1, 1, 0],
                      [1, 1, 1, 1],
                      [1, 1, 1, 1]], dtype=np.int8)
        B = generalized_inverse(A)
        assert ((A @ B @ A) % 2 == A).all()

if __name__ == '__main__':
    unittest.main()


# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file contains circuit simulation routine based on the Fedor's thesis at
# the University of Oxford.
# The method's complexity depends on the cut-rank structure of the ZX diagram.
# Additionally, several heuristics were implemented for finding good rank-decompositions.

import numpy as np
import pyzx_param as zx
from copy import deepcopy
from itertools import product
from typing import Tuple, List, Iterable
from math import sqrt, pi, log2

from .linalg import REF, rank_factorize, generalized_inverse
from .symbolic import Poly


def adjacency_matrix(g: zx.graph.base.BaseGraph, v_left: Iterable[int], v_right: Iterable[int]) -> np.ndarray:
    left_id = {u: i for i, u in enumerate(v_left)}
    right_id = {v: i for i, v in enumerate(v_right)}
    n, m = len(left_id), len(right_id)
    mat = np.zeros((n, m), dtype=np.int8)
    for e in g.edge_set():
        u, v = e[0], e[1]
        if u in left_id and v in right_id:
            mat[left_id[u]][right_id[v]] = 1
        if v in left_id and u in right_id:
            mat[left_id[v]][right_id[u]] = 1
    return mat


def linear_decomposition(arr: Iterable):
    decomp = None
    for elem in arr:
        decomp = elem if decomp is None else [decomp, elem]
    return decomp


def calc_ranks(decomp, g: zx.graph.base.BaseGraph):
    v_all = list(g.vertices())
    v_id = {v: i for i, v in enumerate(v_all)}
    mat = adjacency_matrix(g, v_all, v_all)

    def iterate(elem):
        if isinstance(elem, int):
            ref = REF(mat)
            ref.take(v_id[elem])
            leaves = {v_id[elem]}
            return elem, ref, leaves
        child1, ref1, leaves1 = iterate(elem[0])
        child2, ref2, leaves2 = iterate(elem[1])
        if len(leaves1) >= len(leaves2):
            ref = deepcopy(ref1)
            for j in leaves2:
                ref.take(j)
        else:
            ref = deepcopy(ref2)
            for j in leaves1:
                ref.take(j)
        return [(child1, ref1, leaves1), (child2, ref2, leaves2)], ref, leaves1 | leaves2

    result = iterate(decomp)
    assert result[2] == set(range(g.num_vertices()))
    return result


def rank_width(decomp, g: zx.graph.base.BaseGraph, calc_rs=True) -> int:
    def iterate(data):
        if isinstance(data[0], int):
            return data[1].rank()
        return max(data[1].rank(), iterate(data[0][0]), iterate(data[0][1]))

    if decomp is None:
        return 0
    return iterate(calc_ranks(decomp, g) if calc_rs else decomp)


def rank_score_flops(decomp, g: zx.graph.base.BaseGraph, calc_rs=True) -> float:
    def iterate(data):
        if isinstance(data[0], int):
            return 0
        score1 = iterate(data[0][0])
        score2 = iterate(data[0][1])
        r1, r2, r3 = data[0][0][1].rank(), data[0][1][1].rank(), data[1].rank()
        return score1 + score2 + 2 ** (r1 + r2 + r3 - max(r1, r2, r3))

    if decomp is None:
        return 0
    score = iterate(calc_ranks(decomp, g) if calc_rs else decomp)
    return log2(max(score, 1))


def greedy_linear_order(g: zx.graph.base.BaseGraph) -> List:
    """
    Linear rank-decomposition obtained by greedily taking the next vertex.
    :param g: BaseGraph - ZX diagram
    :return order: list - sequence of vertices for the linear rank-decomposition
    """
    n = g.num_vertices()
    if n == 0:
        return []
    vert = list(g.vertices())
    mat = adjacency_matrix(g, vert, vert)
    ref = REF(mat)
    ref.take(0)
    order = [vert[0]]
    for i in range(1, n):
        min_r, min_u = n, -1
        for u in ref.pivot_cols:
            cur_ref = deepcopy(ref)
            cur_ref.take(u)
            r = cur_ref.rank()
            if r < min_r:
                min_r = r
                min_u = u
        if min_u == -1:
            min_u = ref.taken.index(False)
        order.append(vert[min_u])
        ref.take(min_u)
    return order


def greedy_b2t_decomposition(g: zx.graph.base.BaseGraph):
    """
    Rank-decomposition built greedily from bottom to top.
    :param g: BaseGraph - ZX diagram
    :return decomp: Any - rank-decomposition
    """
    n = g.num_vertices()
    if n == 0:
        return None
    vert = list(g.vertices())
    mat = adjacency_matrix(g, vert, vert)
    refs = []
    for i in range(n):
        ref = REF(mat)
        ref.take(i)
        refs.append(ref)
    decomps = [vert[i] for i in range(n)]
    leaves = [{i} for i in range(n)]
    loc = list(range(n))
    refs_next = dict()
    edges: List = [set() for _ in range(n)]
    for i in range(n):
        for j in refs[i].pivot_cols:
            ref = deepcopy(refs[i])
            ref.take(j)
            refs_next[(i, j)] = refs_next[(j, i)] = ref
            edges[i].add(j)
            edges[j].add(i)

    while refs_next:
        min_rank, i, j = min((ref.rank(), i, j) for (i, j), ref in refs_next.items())
        refs[loc[i]] = refs_next[(i, j)]
        refs.pop(loc[j])
        decomps[loc[i]] = [decomps[loc[i]], decomps[loc[j]]]
        decomps.pop(loc[j])
        leaves[loc[i]] |= leaves[loc[j]]
        leaves.pop(loc[j])
        loc[j] = -1
        for k in range(j + 1, n):
            if loc[k] != -1:
                loc[k] -= 1
        edges_i = edges[i].copy()
        for k in edges_i:
            refs_next.pop((i, k))
            refs_next.pop((k, i))
            edges[i].remove(k)
            edges[k].remove(i)
        edges_j = edges[j].copy()
        for k in edges_j:
            refs_next.pop((j, k))
            refs_next.pop((k, j))
            edges[j].remove(k)
            edges[k].remove(j)
        for k in range(n):
            if loc[k] == -1 or k == i:
                continue
            if (set(refs[loc[i]].pivot_cols) & leaves[loc[k]] or
                    set(refs[loc[k]].pivot_cols) & leaves[loc[i]]):
                i1, i2 = (i, k) if len(leaves[loc[i]]) > len(leaves[loc[k]]) else (k, i)
                ref = deepcopy(refs[loc[i1]])
                for leaf in leaves[loc[i2]]:
                    ref.take(leaf)
                refs_next[(i1, i2)] = refs_next[(i2, i1)] = ref
                edges[i1].add(i2)
                edges[i2].add(i1)

    return linear_decomposition(decomps)


def generate_decomposition(g: zx.graph.base.BaseGraph, strategy='rw-auto', verbose=False):
    """
    Generate rank-decomposition of a ZX diagram using a specific strategy.
    :param g: BaseGraph - ZX diagram
    :param strategy: str - type of strategy
    :param verbose: bool - print additional info
    :return decomp: Any - rank-decomposition
    """
    if strategy == 'rw-greedy-b2t':
        decomp = greedy_b2t_decomposition(g)
    elif strategy == 'rw-greedy-linear':
        decomp = linear_decomposition(greedy_linear_order(g))
    elif strategy == 'rw-auto':
        decomp1 = greedy_b2t_decomposition(g)
        decomp2 = linear_decomposition(greedy_linear_order(g))
        decomps = [decomp1, decomp2]
        scores = [rank_score_flops(decomp, g) for decomp in decomps]
        pos = scores.index(min(scores))
        return decomps[pos]
    else:
        raise ValueError('Unknown rank-decomposition strategy')
    if verbose:
        width, score = rank_width(decomp, g), rank_score_flops(decomp, g)
        print(f'Rank-decomposition ({strategy}) has width {width} and score {score:.3f}')
    return decomp


def mat_image(M: np.ndarray) -> np.ndarray:
    """
    Compute y = xM for all possible vectors x.
    :param M: binary matrix of shape (n, m)
    :return ys: np.ndarray of shape (2^n,) - list of vectors represented as int64
    """
    n, m = M.shape
    M_ints = M @ (1 << np.arange(m))
    ys = np.zeros(1, dtype=np.int64)
    for pos in range(n):
        ys = np.concatenate([ys, M_ints[pos] ^ ys])
    return ys


def apply_parity_map(Psi: np.ndarray, M: np.ndarray) -> np.ndarray:
    """
    Apply parity map M to state Psi.
    :param Psi: np.ndarray of shape (..., 2^n)
    :param M: binary matrix of shape (n, m)
    :return Phi: np.ndarray of shape (..., 2^m)
    """
    m = M.shape[1]
    ys = mat_image(M)
    Phi = np.zeros((Psi.shape[0], 2 ** m), dtype=Psi.dtype)
    np.add.at(Phi, (slice(None), ys), Psi) # type: ignore
    Phi /= sqrt(2) ** (M.sum() - m)
    return Phi


def phase_tensor(E: np.ndarray):
    """
    Generate P_{a,b} = (-1)^{<a, Eb>} / sqrt(2)^|E|.
    :param E: binary matrix of shape (n, m)
    :return P: np.ndarray of shape (2^{n+m},)
    """
    n, m = E.shape
    Eb = mat_image(E.T)
    aEb = np.zeros((1, 2 ** m), dtype=np.int8)
    for pos in range(n):
        aEb = np.vstack([aEb, aEb ^ ((Eb >> pos) & 1)])
    P = (-1) ** aEb.reshape(-1, order='F') / sqrt(2) ** E.sum()
    return P


def conv_naive(Psi_v: np.ndarray, Psi_w: np.ndarray,
               E_vw: np.ndarray, E_vu: np.ndarray, E_wu: np.ndarray) -> np.ndarray:
    """
    Perform convolution naively in 2^{r_u + r_v + r_w} time.
    :param Psi_v: np.ndarray of shape (2^{b_v}, 2^{r_v})
    :param Psi_w: np.ndarray of shape (2^{b_w}, 2^{r_w})
    :param E_vw: binary matrix of shape (r_v, r_w)
    :param E_vu: binary matrix of shape (r_v, r_u)
    :param E_wu: binary matrix of shape (r_w, r_u)
    :return Psi_u: np.ndarray of shape (2^{b_v + b_w}, 2^{r_u})
    """
    r_u, r_v, r_w = E_vu.shape[1], E_vu.shape[0], E_wu.shape[0]
    Psi_v = Psi_v.reshape((-1,) + (2,) * r_v, order='F')
    Psi_w = Psi_w.reshape((-1,) + (2,) * r_w, order='F')
    Psi_u_hat = np.zeros((Psi_v.shape[0], Psi_w.shape[0]) + (2,) * r_u, dtype=Psi_v.dtype)
    for i in range(Psi_v.shape[0]):
        for j in range(Psi_w.shape[0]):
            for x in product(range(2), repeat=r_u):
                for a in product(range(2), repeat=r_v):
                    for b in product(range(2), repeat=r_w):
                        phase = np.dot(a, E_vu @ x) + np.dot(b, E_wu @ x) + np.dot(a, E_vw @ b)
                        Psi_u_hat[i][j][x] += Psi_v[i][a] * Psi_w[j][b] * (-1) ** phase
    Psi_u = np.fft.fftn(Psi_u_hat, axes=tuple(range(2, r_u + 2)))
    Psi_u /= sqrt(2) ** (E_vw.sum() + E_vu.sum() + E_wu.sum() + r_u)
    return Psi_u.reshape((-1, 2 ** r_u), order='F')


def conv_vw(Psi_v: np.ndarray, Psi_w: np.ndarray,
            E_vw: np.ndarray, E_vu: np.ndarray, E_wu: np.ndarray) -> np.ndarray:
    """
    Perform convolution in 2^{r_v + r_w} time:
      1. Compute Psi_v ⊗ Psi_w and multiply by the phase tensor for E_vw
      2. Apply parity map [E_vu; E_wu]
    :param Psi_v: np.ndarray of shape (2^{b_v}, 2^{r_v})
    :param Psi_w: np.ndarray of shape (2^{b_w}, 2^{r_w})
    :param E_vw: binary matrix of shape (r_v, r_w)
    :param E_vu: binary matrix of shape (r_v, r_u)
    :param E_wu: binary matrix of shape (r_w, r_u)
    :return Psi_u: np.ndarray of shape (2^{b_v + b_w}, 2^{r_u})
    """
    Psi_vw = np.kron(Psi_w, Psi_v)
    Psi_vw *= phase_tensor(E_vw)
    E = np.vstack([E_vu, E_wu])
    Psi_u = apply_parity_map(Psi_vw, E)
    return Psi_u


def conv_uv(Psi_v: np.ndarray, Psi_w: np.ndarray,
            E_vw: np.ndarray, E_vu: np.ndarray, E_wu: np.ndarray) -> np.ndarray:
    """
    Perform convolution in 2^{r_u + r_v} time:
      1. Apply parity map [E_vw^T, E_wu] to Psi_w
      2. Multiply by the phase tensor for E_vu
      3. Post-select with Psi_v and apply FT on the second part
    :param Psi_v: np.ndarray of shape (2^{b_v}, 2^{r_v})
    :param Psi_w: np.ndarray of shape (2^{b_w}, 2^{r_w})
    :param E_vw: binary matrix of shape (r_v, r_w)
    :param E_vu: binary matrix of shape (r_v, r_u)
    :param E_wu: binary matrix of shape (r_w, r_u)
    :return Psi_u: np.ndarray of shape (2^{b_v + b_w}, 2^{r_u})
    """
    r_u, r_v, r_w = E_vu.shape[1], E_vu.shape[0], E_wu.shape[0]
    E = np.hstack([E_vw.T, E_wu])
    Psi_vu = apply_parity_map(Psi_w, E).reshape((-1,) + (2,) * r_v + (2 ** r_u,), order='F')
    Psi_vu = (np.fft.fftn(Psi_vu, axes=tuple(range(1, r_v + 1)))
              .reshape((-1, 2 ** (r_v + r_u)), order='F')) / sqrt(2) ** r_v
    E2 = np.block([[np.eye(r_v), E_vu],
                   [np.zeros((r_u, r_v)), np.eye(r_u)]]).astype(np.int8)
    Psi_vu = apply_parity_map(Psi_vu, E2).reshape((-1, 2 ** r_v, 2 ** r_u), order='F')
    Psi_u = np.tensordot(Psi_v, Psi_vu, axes=(1, 1)).reshape((-1, 2 ** r_u), order='F')
    return Psi_u


def conv(Psi_v: np.ndarray, Psi_w: np.ndarray,
         E_vw: np.ndarray, E_vu: np.ndarray, E_wu: np.ndarray, verbose=False) -> Tuple[np.ndarray, bool]:
    """
    Convolution in time O(2^{r_u + r_v + r_w - max(r_u, r_v, r_w)}) by calling a suitable subroutine.
    :param Psi_v: np.ndarray of shape (2^{b_v}, 2^{r_v})
    :param Psi_w: np.ndarray of shape (2^{b_w}, 2^{r_w})
    :param E_vw: binary matrix of shape (r_v, r_w)
    :param E_vu: binary matrix of shape (r_v, r_u)
    :param E_wu: binary matrix of shape (r_w, r_u)
    :param verbose: bool
    :return Psi_u: np.ndarray of shape (2^{b_v + b_w}, 2^{r_u})
            swapped: bool
    """
    r_u, r_v, r_w = E_vu.shape[1], E_vu.shape[0], E_wu.shape[0]
    if verbose:
        print('conv', r_u, r_v, r_w)
    r_max = max(r_u, r_v, r_w)
    if r_u == r_max:
        return conv_vw(Psi_v, Psi_w, E_vw, E_vu, E_wu), False
    elif r_v == r_max:
        return conv_uv(Psi_w, Psi_v, E_vw.T, E_wu, E_vu), True
    else:
        return conv_uv(Psi_v, Psi_w, E_vw, E_vu, E_wu), False


def tensorfy_rw_subtree(g: zx.graph.base.BaseGraph, subtree, verbose=False) -> Tuple[
    np.ndarray, np.ndarray, np.ndarray, List]:
    """
    Returns the simulated version of the subdiagram induced by the leaves from the subtree.
    :param g: BaseGraph - ZX diagram
    :param subtree: Any - subtree of the rank-decomposition
    :param verbose: bool - print additional info
    :return Psi_u: np.ndarray of shape (2^{b_u}, 2^{r_u}) - tensorfication of the subdiagram
            M_u: binary matrix of shape (r_u, n) - parity map to the rest of the graph
            S_u: boolean array - set of vertices in the subdiagram
            B_u: list - order of boundary vertices in the subdiagram
    """
    n = g.num_vertices()
    v_all = np.array(list(g.vertices()))
    v_id = {v: i for i, v in enumerate(g.vertices())}

    if isinstance(subtree, int):
        vt = g.type(subtree)
        if vt == zx.VertexType.BOUNDARY:
            et = g.edge_type(list(g.incident_edges(subtree))[0])
            nb = list(g.neighbors(subtree))[0]
            nbt = g.type(nb)
            is_had = (et == zx.EdgeType.SIMPLE) ^ (nbt == zx.VertexType.BOUNDARY and nb < subtree)
            if is_had:
                Psi_u = np.array([[1, 1], [1, -1]], dtype=np.complex128) / sqrt(2)
            else:
                Psi_u = np.eye(2, dtype=np.complex128)
            B_u = [subtree]
        elif vt == zx.VertexType.Z:
            phase = g.phase(subtree)
            if isinstance(phase, Poly):
                raise ValueError(f"Can't convert diagram with parameters to tensor: {str(phase)}")
            Psi_u = np.array([[1, np.exp(pi * 1j * phase)]])
            B_u = []
        else:
            raise ValueError(f'Invalid vertex type: {vt}')
        M_u = adjacency_matrix(g, [subtree], v_all)
        S_u = np.zeros(n, dtype=bool)
        S_u[v_id[subtree]] = True
        return Psi_u, M_u, S_u, B_u

    Psi_v, M_v, S_v, B_v = tensorfy_rw_subtree(g, subtree[0], verbose=verbose)
    Psi_w, M_w, S_w, B_w = tensorfy_rw_subtree(g, subtree[1], verbose=verbose)
    Cin_v = M_v[:, S_w]
    Cin_w = M_w[:, S_v]
    Cout = np.concatenate((M_v, M_w))
    Cout[:, S_v | S_w] = 0

    r_u, U, V = rank_factorize(Cout)
    r_v, r_w = M_v.shape[0], M_w.shape[0]
    U, V = U[:, :r_u], V[:r_u]
    E_vu, E_wu = U[:r_v], U[r_v:]
    M_u = V

    B = adjacency_matrix(g, v_all[S_v], v_all[S_w])
    Bg = generalized_inverse(B)
    E_vw = (Cin_v @ Bg @ Cin_w.T) % 2
    Psi_u, swapped = conv(Psi_v, Psi_w, E_vw, E_vu, E_wu, verbose=verbose)
    pw = Cout.sum() - U.sum() - V.sum() + r_u
    pw += Cin_v.sum() + Cin_w.sum() - E_vw.sum() - B.sum()
    Psi_u /= sqrt(2) ** pw

    S_u = S_v | S_w
    B_u = B_v + B_w if not swapped else B_w + B_v
    return Psi_u, M_u, S_u, B_u


def tensorfy_rw(g: zx.graph.base.BaseGraph, strategy='rw-auto', preserve_scalar=True, verbose=False) -> np.ndarray:
    """
    Evaluate the tensor diagram corresponding to g using the rank-width method.
    :param g: BaseGraph - ZX diagram
    :param strategy: str - rank-decomposition generation strategy
    :param preserve_scalar: bool - account for the diagram scalar
    :param verbose: bool - print additional info
    :return result: np.ndarray - tensor with |O| + |I| dimensions (output dimensions first)
    """
    g = g.copy()
    zx.full_reduce(g)
    decomp = generate_decomposition(g, strategy=strategy, verbose=verbose)
    if decomp is None:
        return np.array(g.scalar.to_number() ** preserve_scalar)
    result, _, _, boundary = tensorfy_rw_subtree(g, decomp, verbose=verbose)
    result *= g.scalar.to_number() ** preserve_scalar
    order = list(g.outputs()) + list(g.inputs())
    boundary_idx = {v: i for i, v in enumerate(boundary)}
    perm = [boundary_idx[v] for v in order]
    return result.reshape((2,) * len(order), order='F').transpose(perm)



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import math
from fractions import Fraction
from typing import TYPE_CHECKING, Union, Optional, Generic, TypeVar, Any, Sequence
from typing import List, Dict, Set, Tuple, Mapping, Iterable, Callable, ClassVar, Literal
from typing_extensions import Literal, GenericMeta  # type: ignore # https://github.com/python/mypy/issues/5753

import numpy as np

from ..utils import EdgeType, VertexType, toggle_edge, vertex_is_zx, phase_to_s
from ..utils import FloatInt, FractionLike
from ..tensor import tensorfy, tensor_to_matrix

from .scalar import Scalar

if TYPE_CHECKING:
    from .. import simplify


class DocstringMeta(GenericMeta):
    """Metaclass that allows docstring 'inheritance'."""

    def __new__(mcls, classname, bases, cls_dict, **kwargs):
        cls = GenericMeta.__new__(mcls, classname, bases, cls_dict, **kwargs)
        mro = cls.__mro__[1:]
        for name, member in cls_dict.items():
            if not getattr(member, '__doc__'):
                for base in mro:
                    try:
                        member.__doc__ = getattr(base, name).__doc__
                        break
                    except AttributeError:
                        pass
        return cls

def pack_indices(lst: List[FloatInt]) -> Mapping[FloatInt,int]:
    d: Dict[FloatInt,int] = dict()
    if len(lst) == 0: return d
    list.sort(lst)
    i: int = 0
    x: Optional[FloatInt] = None
    for j in range(len(lst)):
        y = lst[j]
        if y != x:
            x = y
            d[y] = i
            i += 1
    return d

VT = TypeVar('VT', bound=int) # The type that is used for representing vertices (e.g. an integer)
ET = TypeVar('ET') # The type used for representing edges (e.g. a pair of integers)


def upair(v1: VT, v2: VT) -> Tuple[VT, VT]:
    """Returns the unordered pair associated to the pair of vertices.
    This method takes a pair of vertices and returns them in a canonical order. Use this
    whenever a pair of vertices is used to reference the location of an undirected edge, 
    e.g. as a key in an edge table."""
    return (v1, v2) if v1 <= v2 else (v2, v1)


class BaseGraph(Generic[VT, ET], metaclass=DocstringMeta):
    """Base class for letting graph backends interact with PyZX.
    For a backend to work with PyZX, there should be a class that implements
    all the methods of this class. For implementations of this class see
    :class:`~pyzx.graph.graph_s.GraphS` or :class:`~pyzx.graph.graph_ig.GraphIG`."""

    backend: ClassVar[str] = 'None'

    def __init__(self) -> None:
        self.scalar: Scalar = Scalar()
        # Default storage for symbolic phases; backends can override these in their own __init__.
        self._phase: Dict[VT, FractionLike] = dict()
        self._phaseVars: Dict[VT, Set[str]] = dict()
        # self.inputs: List[VT] = []
        # self.outputs: List[VT] = []
        #Data necessary for phase tracking for phase teleportation
        self.track_phases: bool = False
        self.phase_index : Dict[VT,int] = dict() # {vertex:index tracking its phase for phase teleportation}
        self.phase_master: Optional['simplify.Simplifier'] = None
        self.phase_mult: Dict[int,Literal[1,-1]] = dict()
        self.max_phase_index: int = -1

        # merge_vdata(v0,v1) is an optional, custom function for merging
        # vdata of v1 into v0 during spider fusion etc.
        self.merge_vdata: Optional[Callable[[VT,VT], None]] = None
        self.variable_types: Dict[str,bool] = dict() # mapping of variable names to their type (bool or continuous)

    # MANDATORY OVERRIDES {{{

    # All backends should override these methods

    def clone(self) -> BaseGraph[VT,ET]:
        """
        This method should return an identical copy of the graph, without any relabeling.

        This needs to be implemented in the backend, since different backends deal with names differently.
        """
        raise NotImplementedError()

    def inputs(self) -> Tuple[VT, ...]:
        """Gets the inputs of the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def set_inputs(self, inputs: Tuple[VT, ...]):
        """Sets the inputs of the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def outputs(self) -> Tuple[VT, ...]:
        """Gets the outputs of the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def set_outputs(self, outputs: Tuple[VT, ...]):
        """Sets the outputs of the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def add_vertices(self, amount: int) -> List[VT]:
        """Add the given amount of vertices, and return the indices of the
        new vertices added to the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def add_vertex_indexed(self,v:VT) -> None:
        """Adds a vertex that is guaranteed to have the chosen index (i.e. 'name').
        If the index isn't available, raises a ValueError.
        This method is used in the editor and ZXLive to support undo,
        which requires vertices to preserve their index."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def add_edge(self, edge_pair: Tuple[VT,VT], edgetype:EdgeType=EdgeType.SIMPLE) -> ET:
        """Adds a single edge of the given type and return its id"""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def remove_vertices(self, vertices: Iterable[VT]) -> None:
        """Removes the list of vertices from the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def remove_edges(self, edges: List[ET]) -> None:
        """Removes the list of edges from the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def num_vertices(self) -> int:
        """Returns the amount of vertices in the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def num_edges(self, s: Optional[VT]=None, t: Optional[VT]=None, et: Optional[EdgeType]=None) -> int:
        """Returns the amount of edges in the graph"""
        return len(list(self.edges(s, t)))

    def vertices(self) -> Iterable[VT]:
        """Iterator over all the vertices."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def edges(self, s: Optional[VT]=None, t: Optional[VT]=None) -> Iterable[ET]:
        """Iterator that returns all the edges in the graph, or all the edges connecting the pair of vertices.
        Output type depends on implementation in backend."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def edge_st(self, edge: ET) -> Tuple[VT, VT]:
        """Returns a tuple of source/target of the given edge."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def incident_edges(self, vertex: VT) -> Sequence[ET]:
        """Returns all neighboring edges of the given vertex."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def edge_type(self, e: ET) -> EdgeType:
        """Returns the type of the given edge:
        ``EdgeType.SIMPLE`` if it is regular, ``EdgeType.HADAMARD`` if it is a Hadamard edge,
        0 if the edge is not in the graph."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def set_edge_type(self, e: ET, t: EdgeType) -> None:
        """Sets the type of the given edge."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def type(self, vertex: VT) -> VertexType:
        """Returns the type of the given vertex:
        VertexType.BOUNDARY if it is a boundary, VertexType.Z if it is a Z node,
        VertexType.X if it is a X node, VertexType.H_BOX if it is an H-box."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def set_type(self, vertex: VT, t: VertexType) -> None:
        """Sets the type of the given vertex to t."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def phase(self, vertex: VT) -> FractionLike:
        """Returns the phase value of the given vertex."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def set_phase(self, vertex: VT, phase: FractionLike) -> None:
        """Sets the phase of the vertex to the given value."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def qubit(self, vertex: VT) -> FloatInt:
        """Returns the qubit index associated to the vertex.
        If no index has been set, returns -1."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def set_qubit(self, vertex: VT, q: FloatInt) -> None:
        """Sets the qubit index associated to the vertex."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def row(self, vertex: VT) -> FloatInt:
        """Returns the row that the vertex is positioned at.
        If no row has been set, returns -1."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def set_row(self, vertex: VT, r: FloatInt) -> None:
        """Sets the row the vertex should be positioned at."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def clear_vdata(self, vertex: VT) -> None:
        """Removes all vdata associated to a vertex"""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def vdata_keys(self, vertex: VT) -> Sequence[str]:
        """Returns an iterable of the vertex data key names.
        Used e.g. in making a copy of the graph in a backend-independent way."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def vdata(self, vertex: VT, key: str, default: Any=None) -> Any:
        """Returns the data value of the given vertex associated to the key.
        If this key has no value associated with it, it returns the default value."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def set_vdata(self, vertex: VT, key: str, val: Any) -> None:
        """Sets the vertex data associated to key to val."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def clear_edata(self, edge: ET) -> None:
        """Removes all edata associated to an edge"""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def edata_keys(self, edge: ET) -> Sequence[str]:
        """Returns an iterable of the edge data key names."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def edata(self, edge: ET, key: str, default: Any=None) -> Any:
        """Returns the data value of the given edge associated to the key.
        If this key has no value associated with it, it returns the default value."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def set_edata(self, edge: ET, key: str, val: Any) -> None:
        """Sets the edge data associated to key to val."""
        raise NotImplementedError("Not implemented on backend " + type(self).backend)
    # }}}


    # OPTIONAL OVERRIDES {{{

    # These only need to be overridden if the backend will be used with hybrid classical/quantum
    # methods.
    def is_ground(self, vertex: VT) -> bool:
        """Returns a boolean indicating if the vertex is connected to a ground."""
        return False

    def grounds(self) -> Set[VT]:
        """Returns the set of vertices connected to a ground."""
        return set(v for v in self.vertices() if self.is_ground(v))

    def set_ground(self, vertex: VT, flag: bool=True) -> None:
        """Connect or disconnect the vertex to a ground."""
        raise NotImplementedError("Not implemented on backend" + type(self).backend)

    def is_hybrid(self) -> bool:
        """Returns whether this is a hybrid quantum-classical graph,
        i.e. a graph with ground generators."""
        return bool(self.grounds())

    # Override and set to true if the backend supports parallel edges
    def multigraph(self) -> bool:
        return False


    # Backends may wish to override these methods to implement them more efficiently

    # These methods return mappings from vertices to various pieces of data. If the backend
    # stores these e.g. as Python dicts, just return the relevant dicts.
    def phases(self) -> Mapping[VT, FractionLike]:
        """Returns a mapping of vertices to their phase values."""
        return { v: self.phase(v) for v in self.vertices() }

    def types(self) -> Mapping[VT, VertexType]:
        """Returns a mapping of vertices to their types."""
        return { v: self.type(v) for v in self.vertices() }

    def qubits(self) -> Mapping[VT,FloatInt]:
        """Returns a mapping of vertices to their qubit index."""
        return { v: self.qubit(v) for v in self.vertices() }

    def rows(self) -> Mapping[VT, FloatInt]:
        """Returns a mapping of vertices to their row index."""
        return { v: self.row(v) for v in self.vertices() }

    def depth(self) -> FloatInt:
        """Returns the value of the highest row number given to a vertex.
        This is -1 when no rows have been set."""
        if self.num_vertices() == 0:
            return -1
        else:
            return max(self.row(v) for v in self.vertices())

    def edge(self, s:VT, t:VT, et: EdgeType=EdgeType.SIMPLE) -> ET:
        """Returns the name of the first edge with the given source/target and type. Behaviour is undefined if the vertices are not connected."""
        for e in self.incident_edges(s):
            if t in self.edge_st(e) and et == self.edge_type(e):
                return e
        raise ValueError(f"No edge of type {et} between {s} and {t}")

    def connected(self,v1: VT,v2: VT) -> bool:
        """Returns whether vertices v1 and v2 share an edge."""
        for e in self.incident_edges(v1):
            if v2 in self.edge_st(e):
                return True
        return False

    def add_vertex(self,
                   ty:VertexType=VertexType.BOUNDARY,
                   qubit:FloatInt=-1,
                   row:FloatInt=-1,
                   phase:Optional[FractionLike]=None,
                   ground:bool=False,
                   index: Optional[VT] = None,
                   phaseVars: Optional[Set[str]] = None
                   ) -> VT:
        """Add a single vertex to the graph and return its index.
        The optional parameters allow you to respectively set
        the type, qubit index, row index and phase of the vertex."""
        if index is not None:
            self.add_vertex_indexed(index)
            v = index
        else:
            v = self.add_vertices(1)[0]
        self.set_type(v, ty)
        if phase is None:
            if ty == VertexType.H_BOX: phase = 1
            else: phase = 0
        self.set_qubit(v, qubit)
        self.set_row(v, row)
        if phase:
            self.set_phase(v, phase)
        if phaseVars is not None:
            self.set_params(v, phaseVars)
        if ground:
            self.set_ground(v, True)
        if self.track_phases:
            self.max_phase_index += 1
            self.phase_index[v] = self.max_phase_index
            self.phase_mult[self.max_phase_index] = 1
        return v

    def add_edges(self, edge_pairs: Iterable[Tuple[VT,VT]], edgetype:EdgeType=EdgeType.SIMPLE) -> None:
        """Adds a list of edges to the graph."""
        for ep in edge_pairs:
            self.add_edge(ep, edgetype)

    def remove_vertex(self, vertex: VT) -> None:
        """Removes the given vertex from the graph."""
        self.remove_vertices([vertex])

    def remove_edge(self, edge: ET) -> None:
        """Removes the given edge from the graph."""
        self.remove_edges([edge])

    def get_all_params(self) -> Mapping[VT, Set[str]]:
        """Returns a mapping of vertices to their symbolic phase parameters."""
        return self._phaseVars

    def get_params(self, vertex: VT) -> Set[str]:
        """Returns the symbolic parameters attached to the given vertex."""
        return self._phaseVars.get(vertex, set())

    def set_params(self, vertex: VT, params: Set[str]) -> None:
        """Sets the symbolic parameters attached to the given vertex."""
        self._phaseVars[vertex] = set(params)

    def add_params(self, vertex: VT, params: Set[str]) -> None:
        """Adds (XORs) the given symbolic parameters to the vertex."""
        current = self._phaseVars.get(vertex, set())
        self._phaseVars[vertex] = current.union(params)

    def add_to_phase(self, vertex: VT, phase: FractionLike, params: Optional[Set[str]] = None) -> None:
        """Add the given phase and optional symbolic params to the given vertex."""
        params = params or set()
        if params:
            self.add_params(vertex, params)
        self.set_phase(vertex, self.phase(vertex)+phase)

    def get_phase_str(self, vertex: VT) -> str:
        """Human readable representation of the phase and params of a vertex."""
        return phase_to_s(self.phase(vertex), self.type(vertex))

    def num_inputs(self) -> int:
        """Gets the number of inputs of the graph."""
        return len(self.inputs())

    def num_outputs(self) -> int:
        """Gets the number of outputs of the graph."""
        return len(self.outputs())

    def set_position(self, vertex: VT, q: FloatInt, r: FloatInt):
        """Set both the qubit index and row index of the vertex."""
        self.set_qubit(vertex, q)
        self.set_row(vertex, r)

    def neighbors(self, vertex: VT) -> Sequence[VT]:
        """Returns all neighboring vertices of the given vertex."""
        vs: Set[VT] = set()
        for e in self.incident_edges(vertex):
            s,t = self.edge_st(e)
            vs.add(s if t == vertex else t)
        return list(vs)

    def vertex_degree(self, vertex: VT) -> int:
        """Returns the degree of the given vertex."""
        return len(self.incident_edges(vertex))

    def edge_s(self, edge: ET) -> VT:
        """Returns the source of the given edge."""
        return self.edge_st(edge)[0]

    def edge_t(self, edge: ET) -> VT:
        """Returns the target of the given edge."""
        return self.edge_st(edge)[1]


    def vertex_set(self) -> Set[VT]:
        """Returns the vertices of the graph as a Python set.
        Should be overloaded if the backend supplies a cheaper version than this."""
        return set(self.vertices())

    def edge_set(self) -> Set[ET]:
        """Returns the edges of the graph as a Python set.
        Should be overloaded if the backend supplies a cheaper version than this. Note this ignores parallel edges."""
        return set(self.edges())


    # }}}

    # def vindex(self) -> VT:
    #     """The index given to the next vertex added to the graph. It should always
    #     be equal to ``max(g.vertices()) + 1``."""
    #     raise NotImplementedError("Not implemented on backend " + type(self).backend)

    def __str__(self) -> str:
        return "Graph({} vertices, {} edges)".format(
                str(self.num_vertices()),str(self.num_edges()))

    def __repr__(self) -> str:
        return str(self)

    def stats(self) -> str:
        """
        Returns:
            Returns a string with some information regarding the degree distribution of the graph.
        """
        s = str(self) + "\n"
        degrees: Dict[int,int] = {}
        for v in self.vertices():
            d = self.vertex_degree(v)
            if d in degrees: degrees[d] += 1
            else: degrees[d] = 1
        s += "degree distribution: \n"
        for d, n in sorted(degrees.items(),key=lambda x: x[0]):
            s += "{:d}: {:d}\n".format(d,n)
        return s

    def copy(self, adjoint:bool=False, backend:Optional[str]=None) -> BaseGraph[VT,ET]:
        """Create a copy of the graph. If ``adjoint`` is set,
        the adjoint of the graph will be returned (inputs and outputs flipped, phases reversed).
        When ``backend`` is set, a copy of the graph with the given backend is produced.
        By default the copy will have the same backend.

        Args:
            adjoint: set to True to make the copy be the adjoint of the graph
            backend: the backend of the output graph

        Returns:
            A copy of the graph

        Note:
            The copy will have consecutive vertex indices, even if the original
            graph did not.
        """
        from .graph import Graph # imported here to prevent circularity
        from .multigraph import Multigraph
        if (backend is None):
            backend = type(self).backend
        g = Graph(backend = backend)
        if isinstance(self, Multigraph) and isinstance(g, Multigraph):
            g.set_auto_simplify(self._auto_simplify) # type: ignore
            # mypy issue https://github.com/python/mypy/issues/16413
        g.track_phases = self.track_phases
        g.scalar = self.scalar.copy(conjugate=adjoint)
        g.merge_vdata = self.merge_vdata # type: ignore
        mult:int = 1
        if adjoint:
            mult = -1

        #g.add_vertices(self.num_vertices())
        ty = self.types()
        ph = self.phases()
        pvs = self.get_all_params() if hasattr(self, "get_all_params") else {}
        qs = self.qubits()
        rs = self.rows()
        maxr = self.depth()
        vtab = dict()
        for v in self.vertices():
            pv = pvs[v] if v in pvs else set()
            i = g.add_vertex(ty[v], phase=mult * ph[v], phaseVars=pv)
            if v in qs: g.set_qubit(i,qs[v])
            if v in rs:
                if adjoint: g.set_row(i, maxr-rs[v])
                else: g.set_row(i, rs[v])
            vtab[v] = i
            for k in self.vdata_keys(v):
                g.set_vdata(i, k, self.vdata(v, k))
        for v in self.grounds():
            g.set_ground(vtab[v], True)

        new_inputs = tuple(vtab[i] for i in self.inputs())
        new_outputs = tuple(vtab[i] for i in self.outputs())
        if not adjoint:
            g.set_inputs(new_inputs)
            g.set_outputs(new_outputs)
        else:
            g.set_inputs(new_outputs)
            g.set_outputs(new_inputs)
        
        for e in self.edges():
            s, t = self.edge_st(e)
            new_e = g.add_edge((vtab[s], vtab[t]), self.edge_type(e))
            g.set_edata_dict(new_e, self.edata_dict(e))

        return g

    def adjoint(self) -> BaseGraph[VT,ET]:
        """Returns a new graph equal to the adjoint of this graph."""
        return self.copy(adjoint=True)


    def map_qubits(self, qubit_map:Mapping[int,Tuple[float,float]]) -> None:
        for v in self.vertices():
            q = self.qubit(v)
            r = self.row(v)
            q0 = min(max(0,math.floor(q)), len(qubit_map)-1)
            offset = q - q0
            coord = qubit_map[q0]
            qf = 3*(coord[0]+offset)+(0.6 * coord[1])
            rf = 3*r+(0.6 * coord[1])
            self.set_qubit(v, qf)
            self.set_row(v, rf)

    def compose(self, other: BaseGraph[VT,ET]) -> None:
        """Inserts a graph after this one. The amount of qubits of the graphs must match.
        Also available by the operator `graph1 + graph2`"""
        other = other.copy()
        outputs = self.outputs()
        inputs = other.inputs()
        if len(outputs) != len(inputs):
            raise TypeError("Outputs of first graph must match inputs of second.")

        plugs: List[Tuple[VT,VT,EdgeType]] = []
        for k in range(len(outputs)):
            o = outputs[k]
            i = inputs[k]
            if len(self.neighbors(o)) != 1:
                raise ValueError("Bad output vertex: " + str(o))
            if len(other.neighbors(i)) != 1:
                raise ValueError("Bad input vertex: " + str(i))
            no = next(iter(self.neighbors(o)))
            ni = next(iter(other.neighbors(i)))
            plugs.append((no, ni, EdgeType.HADAMARD
                if self.edge_type(self.edge(no,o)) != other.edge_type(other.edge(i,ni))
                else EdgeType.SIMPLE))

        self.scalar.mult_with_scalar(other.scalar)
        maxr = max((self.row(v) for v in self.vertices()
                    if self.type(v) != VertexType.BOUNDARY), default=0)
        minr = min((other.row(v) for v in other.vertices()
                    if other.type(v) != VertexType.BOUNDARY), default=0)
        offset = maxr - minr + 1

        for v in outputs:
            self.remove_vertex(v)

        vtab : Dict[VT,VT] = dict()
        for v in other.vertices():
            if not v in inputs:
                w = self.add_vertex(other.type(v),
                        phase=other.phase(v),
                        qubit=other.qubit(v),
                        row=offset + other.row(v),
                        ground=other.is_ground(v))
                self.set_vdata_dict(w, other.vdata_dict(v))
                self.set_params(w, other.get_params(v))
                vtab[v] = w
        for e in other.edges():
            s,t = other.edge_st(e)
            if not s in inputs and not t in inputs:
                new_e = self.add_edge((vtab[s],vtab[t]), edgetype=other.edge_type(e))
                self.set_edata_dict(new_e, other.edata_dict(e))

        for (no,ni,et) in plugs:
            self.add_edge((no,vtab[ni]), edgetype=et)
        self.set_outputs(tuple(vtab[v] for v in other.outputs()))

    def tensor(self, other: BaseGraph[VT,ET]) -> BaseGraph[VT,ET]:
        """Take the tensor product of two graphs. Places the second graph below the first one.
        Can also be called using the operator ``graph1 @ graph2``"""
        g = self.copy()
        g.scalar.mult_with_scalar(other.scalar)
        ts = other.types()
        qs = other.qubits()
        height = max((self.qubits().values()), default=0) + 1
        rs = other.rows()
        phases = other.phases()
        vertex_map = dict()
        for v in other.vertices():
            w = g.add_vertex(ts[v],qs[v]+height,rs[v],phases[v],g.is_ground(v))
            g.set_vdata_dict(w, other.vdata_dict(v))
            vertex_map[v] = w
        for e in other.edges():
            s,t = other.edge_st(e)
            new_e = g.add_edge((vertex_map[s],vertex_map[t]),other.edge_type(e))
            g.set_edata_dict(new_e, other.edata_dict(e))

        inputs = g.inputs() + tuple(vertex_map[v] for v in other.inputs())
        outputs = g.outputs() + tuple(vertex_map[v] for v in other.outputs())
        g.set_inputs(inputs)
        g.set_outputs(outputs)

        minr = min((g.row(v) for v in inputs), default=1)
        for v in inputs:
            g.set_row(v, minr)

        maxr = max((g.row(v) for v in outputs), default=1)
        for v in outputs:
            g.set_row(v, maxr)

        return g

    def __iadd__(self, other: BaseGraph[VT,ET]) -> BaseGraph[VT,ET]:
        self.compose(other)
        return self

    def __add__(self, other: BaseGraph[VT,ET]) -> BaseGraph[VT,ET]:
        g = self.copy()
        g += other
        return g

    def __mul__(self, other: BaseGraph[VT,ET]) -> BaseGraph[VT,ET]:
        """Compose two diagrams, in formula order. That is, g * h produces 'g AFTER h'."""
        g = other.copy()
        g.compose(self)
        return g

    def __matmul__(self, other: BaseGraph[VT,ET]) -> BaseGraph[VT,ET]:
        return self.tensor(other)

    def merge(self, other: BaseGraph[VT,ET]) -> Tuple[List[VT],List[ET]]:
        """Merges this graph with the other graph in-place.
        Returns (list-of-vertices, list-of-edges) corresponding to
        the id's of the vertices and edges of the other graph."""
        ty = other.types()
        rs = other.rows()
        qs = other.qubits()
        phase = other.phases()
        grounds = other.grounds()
        phasevars = other.get_all_params()

        vert_map = dict()
        edges = []
        for v in other.vertices():
            w = self.add_vertex(ty[v], qs[v], rs[v], phase[v], v in grounds, phaseVars=phasevars.get(v, set()))
            self.set_vdata_dict(w, other.vdata_dict(v))
            vert_map[v] = w
        for e in other.edges():
            s,t = other.edge_st(e)
            f = (vert_map[s],vert_map[t])
            new_e = self.add_edge(f,other.edge_type(e))
            self.set_edata_dict(new_e, other.edata_dict(e))
            edges.append(e)
        return (list(vert_map.values()),edges)

    def subgraph_from_vertices(self,verts: List[VT]) -> BaseGraph[VT,ET]:
        """Returns the subgraph consisting of the specified vertices."""
        from .graph import Graph # imported here to prevent circularity
        from .multigraph import Multigraph
        g = Graph(backend=type(self).backend)
        if isinstance(self, Multigraph) and isinstance(g, Multigraph):
            g.set_auto_simplify(self._auto_simplify) # type: ignore
            # mypy issue https://github.com/python/mypy/issues/16413
        ty = self.types()
        rs = self.rows()
        qs = self.qubits()
        phase = self.phases()
        grounds = self.grounds()
        phasevars = self.get_all_params()

        edges = [e for e in self.edges() \
            if self.edge_st(e)[0] in verts and self.edge_st(e)[1] in verts]

        vert_map = dict()
        for v in verts:
            w = g.add_vertex(ty[v], qs[v], rs[v], phase[v], v in grounds, index=v, phaseVars=phasevars.get(v, set()))
            vert_map[v] = w
            g.set_vdata_dict(w, self.vdata_dict(v))
        for e in edges:
            s,t = self.edge_st(e)
            new_e = g.add_edge((vert_map[s], vert_map[t]), self.edge_type(e))
            g.set_edata_dict(new_e, self.edata_dict(e))

        return g

    def apply_state(self, state: str) -> None:
        """Inserts a state into the inputs of the graph. ``state`` should be
        a string with every character representing an input state for each qubit.
        The possible types of states are on of '0', '1', '+', '-' for the respective
        kets. If '/' is specified this input is skipped."""
        inputs = self.inputs()
        if len(state) > len(inputs): raise TypeError("Too many input states specified")
        new_inputs = []
        for i,s in enumerate(state):
            v = inputs[i]
            if s == '/':
                new_inputs.append(v)
                continue
            if s in ('0', '1'):
                self.scalar.add_power(-1)
                self.set_type(v, VertexType.X)
                if s == '1':
                    self.set_phase(v, Fraction(1))
            elif s in ('+', '-'):
                self.scalar.add_power(-1)
                self.set_type(v, VertexType.Z)
                if s == '-':
                    self.set_phase(v, Fraction(1))
            else:
                raise TypeError("Unknown input state " + s)
        self.set_inputs(tuple(new_inputs))

    def apply_effect(self, effect: str) -> None:
        """Inserts an effect into the outputs of the graph. ``effect`` should be
        a string with every character representing an output effect for each qubit.
        The possible types of effects are one of '0', '1', '+', '-' for the respective
        kets. If '/' is specified this output is skipped."""
        outputs = self.outputs()
        if len(effect) > len(outputs): raise TypeError("Too many output effects specified")
        new_outputs = []
        for i,s in enumerate(effect):
            v = outputs[i]
            if s == '/': 
                new_outputs.append(v)
                continue
            if s in ('0', '1'):
                self.scalar.add_power(-1)
                self.set_type(v, VertexType.X)
                if s == '1':
                    self.set_phase(v, Fraction(1))
            elif s in ('+', '-'):
                self.scalar.add_power(-1)
                self.set_type(v, VertexType.Z)
                if s == '-':
                    self.set_phase(v, Fraction(1))
            else:
                raise TypeError("Unknown output effect " + s)
        self.set_outputs(tuple(new_outputs))

    def to_tensor(self, preserve_scalar:bool=True) -> np.ndarray:
        """Returns a representation of the graph as a tensor using :func:`~pyzx.tensor.tensorfy`"""
        return tensorfy(self, preserve_scalar)
    def to_matrix(self,preserve_scalar:bool=True) -> np.ndarray:
        """Returns a representation of the graph as a matrix using :func:`~pyzx.tensor.tensorfy`"""
        return tensor_to_matrix(tensorfy(self, preserve_scalar), self.num_inputs(), self.num_outputs())

    def to_dict(self, include_scalar:bool=True) -> Dict[str, Any]:
        """Returns a dictionary representation of the graph, which can then be converted into json."""
        from .jsonparser import graph_to_dict
        return graph_to_dict(self, include_scalar)

    def to_json(self, include_scalar:bool=True) -> str:
        """Returns a json representation of the graph.
        Convert back into a graph using :meth:`from_json`."""
        from .jsonparser import graph_to_json
        return graph_to_json(self, include_scalar)

    def to_graphml(self) -> str:
        """Returns a GraphML representation of the graph."""
        from .jsonparser import to_graphml
        return to_graphml(self)

    def to_tikz(self,draw_scalar:bool=False) -> str:
        """Returns a Tikz representation of the graph."""
        from ..tikz import to_tikz
        return to_tikz(self,draw_scalar)

    @classmethod
    def from_json(cls, js:Union[str,Dict[str,Any]]) -> BaseGraph[VT,ET]:
        """Converts the given .qgraph json string into a Graph.
        Works with the output of :meth:`to_json`."""
        from .jsonparser import json_to_graph
        return json_to_graph(js)

    @classmethod
    def from_tikz(cls, tikz: str, warn_overlap:bool= True, fuse_overlap:bool = True, ignore_nonzx:bool = False) -> BaseGraph[VT,ET]:
        """Converts a tikz diagram into a pyzx Graph.
    The tikz diagram is assumed to be one generated by Tikzit,
    and hence should have a nodelayer and a edgelayer..

    Args:
        s: a string containing a well-defined Tikz diagram.
        warn_overlap: If True raises a Warning if two vertices have the exact same position.
        fuse_overlap: If True fuses two vertices that have the exact same position. Only has effect if fuse_overlap is False.
        ignore_nonzx: If True suppresses most errors about unknown vertex/edge types and labels.

    Warning:
        Vertices that might look connected in the output of the tikz are not necessarily connected
        at the level of tikz itself, and won't be treated as such in pyzx.
    """
        from ..tikz import tikz_to_graph
        return tikz_to_graph(tikz,warn_overlap, fuse_overlap, ignore_nonzx, cls.backend)

    def is_id(self) -> bool:
        """Returns whether the graph is just a set of identity wires,
        i.e. a graph where all the vertices are either inputs or outputs,
        and they are connected to each other in a non-permuted manner."""
        inputs = self.inputs()
        outputs = self.outputs()

        if (len(inputs) != len(outputs) or
            self.num_vertices() != 2*len(inputs) or
            self.num_edges() != len(inputs)): return False

        for i in range(len(inputs)):
            if not self.connected(inputs[i], outputs[i]): return False
        return True

    def pack_circuit_rows(self) -> None:
        """Compresses the rows of the graph so that every index is used."""
        rows = [self.row(v) for v in self.vertices()]
        new_rows = pack_indices(rows)
        for v in self.vertices():
            self.set_row(v, new_rows[self.row(v)])

    def qubit_count(self) -> int:
        """Returns the number of inputs of the graph"""
        return self.num_inputs()

    def auto_detect_io(self):
        """Adds every vertex that is of boundary-type to the list of inputs or outputs.
        Whether it is an input or output is determined by looking whether its neighbor
        is further to the right or further to the left of the input.
        Inputs and outputs are sorted by vertical position.
        Raises an exception if boundary vertex does not have a unique neighbor
        or if this neighbor is on the same horizontal position.
        """
        ty = self.types()
        inputs = []
        outputs = []
        for v in self.vertices():
            if ty[v] != VertexType.BOUNDARY: continue
            if v in inputs or v in outputs: continue
            if self.vertex_degree(v) != 1:
                raise TypeError("Invalid ZX-diagram: Boundary-type vertex does not have unique neighbor")
            w = list(self.neighbors(v))[0]
            if self.row(w) > self.row(v):
                inputs.append(v)
            elif self.row(w) < self.row(v):
                outputs.append(v)
            else:
                raise TypeError("Boundary-type vertex at same horizontal position as neighbor. Can't determine whether it is an input or output.")
        inputs = sorted(filter(lambda v: v in self.vertices(), inputs), key=self.qubit)
        outputs = sorted(filter(lambda v: v in self.vertices(), outputs), key=self.qubit)
        self.set_inputs(tuple(inputs))
        self.set_outputs(tuple(outputs))

    def normalize(self) -> None:
        """Puts every node connecting to an input/output at the correct qubit index and row."""
        if self.num_inputs() == 0:
            self.auto_detect_io()
        max_r = self.depth() - 1
        if max_r <= 2:
            for o in self.outputs():
                self.set_row(o,4)
            max_r = self.depth() -1
        claimed = []
        for q,i in enumerate(sorted(self.inputs(), key=self.qubit)):
            self.set_row(i,0)
            self.set_qubit(i,q)
            #q = self.qubit(i)
            n = list(self.neighbors(i))[0]
            if self.type(n) in (VertexType.Z, VertexType.X):
                claimed.append(n)
                self.set_row(n,1)
                self.set_qubit(n, q)
            else: #directly connected to output
                e = self.edge(i, n)
                t = self.edge_type(e)
                self.remove_edge(e)
                v = self.add_vertex(VertexType.Z,q,1)
                self.add_edge((i,v),toggle_edge(t))
                self.add_edge((v,n),EdgeType.HADAMARD)
                claimed.append(v)
        for q, o in enumerate(sorted(self.outputs(),key=self.qubit)):
            #q = self.qubit(o)
            self.set_row(o,max_r+1)
            self.set_qubit(o,q)
            n = list(self.neighbors(o))[0]
            if n not in claimed:
                self.set_row(n,max_r)
                self.set_qubit(n, q)
            else:
                e = self.edge(o, n)
                t = self.edge_type(e)
                self.remove_edge(e)
                v = self.add_vertex(VertexType.Z,q,max_r)
                self.add_edge((o,v),toggle_edge(t))
                self.add_edge((v,n),EdgeType.HADAMARD)

        self.pack_circuit_rows()

    def translate(self, x:FloatInt, y:FloatInt) -> BaseGraph[VT,ET]:
        g = self.copy()
        for v in g.vertices():
            g.set_row(v, g.row(v)+x)
            g.set_qubit(v,g.qubit(v)+y)
        return g



    def add_edge_table(self, etab:Mapping[Tuple[VT,VT],List[int]]) -> None:
        """Takes a dictionary mapping (source,target) --> (#edges, #h-edges) specifying that
        #edges regular edges must be added between source and target and $h-edges Hadamard edges.
        The method selectively adds or removes edges to produce that ZX diagram which would
        result from adding (#edges, #h-edges), and then removing all parallel edges using Hopf/spider laws."""

        for st, (ns, nh) in etab.items():
            for _ in range(ns): self.add_edge(st, EdgeType.SIMPLE)
            for _ in range(nh): self.add_edge(st, EdgeType.HADAMARD)


    def set_phase_master(self, m: 'simplify.Simplifier') -> None:
        """Points towards an instance of the class :class:`~pyzx.simplify.Simplifier`.
        Used for phase teleportation."""
        self.phase_master = m

    def update_phase_index(self, old:VT, new:VT) -> None:
        """When a phase is moved from a vertex to another vertex,
        we need to tell the phase_teleportation algorithm that this has happened.
        This function does that. Used in some of the rules in `simplify`."""
        if not self.track_phases: return
        i = self.phase_index[old]
        self.phase_index[old] = self.phase_index[new]
        self.phase_index[new] = i

    def fuse_phases(self, p1: VT, p2: VT) -> None:
        if p1 not in self.phase_index or p2 not in self.phase_index:
            return
        if self.phase_master is not None:
            self.phase_master.fuse_phases(self.phase_index[p1],self.phase_index[p2])
        self.phase_index[p2] = self.phase_index[p1]

    def phase_negate(self, v: VT) -> None:
        if v not in self.phase_index: return
        index = self.phase_index[v]
        mult = self.phase_mult[index]
        if mult == 1: self.phase_mult[index] = -1
        else: self.phase_mult[index] = 1
        #self.phase_mult[index] = -1*mult

    def vertex_from_phase_index(self, i: int) -> VT:
        return list(self.phase_index.keys())[list(self.phase_index.values()).index(i)]


    def remove_isolated_vertices(self) -> None:
        """Deletes all vertices and vertex pairs that are not connected to any other vertex."""
        rem: List[VT] = []
        for v in self.vertices():
            d = self.vertex_degree(v)
            phase = self.phase(v)
            if phase.denominator not in [1, 2, 4]:
                # TODO: allow evaluation of isolated non-Clifford+T vertices
                continue
            if d == 0:
                rem.append(v)
                ty = self.type(v)
                if ty == VertexType.BOUNDARY:
                    raise TypeError("Diagram is not a well-typed ZX-diagram: contains isolated boundary vertex.")
                elif ty == VertexType.H_BOX:
                    self.scalar.add_phase(self.phase(v))
                else: self.scalar.add_node(self.phase(v), self.get_params(v))
            if d == 1: # It has a unique neighbor
                if v in rem: continue # Already taken care of
                if self.type(v) == VertexType.BOUNDARY: continue # Ignore in/outputs
                w = list(self.neighbors(v))[0]
                if len(list(self.neighbors(w))) > 1: continue # But this neighbor has other neighbors
                if self.type(w) == VertexType.BOUNDARY: continue # It's a state/effect
                if self.phase(w).denominator not in [1, 2, 4]:
                    continue
                # At this point w and v are only connected to each other
                rem.append(v)
                rem.append(w)
                et = self.edge_type(self.edge(v,w))
                t1 = self.type(v)
                t2 = self.type(w)
                params_vw = self.get_params(v).symmetric_difference(self.get_params(w))
                if t1 == VertexType.H_BOX: t1 = VertexType.Z # 1-ary H-box is just a Z spider
                if t2 == VertexType.H_BOX: t2 = VertexType.Z
                if t1==t2:
                    if et == EdgeType.SIMPLE:
                        self.scalar.add_node(self.phase(v)+self.phase(w), params_vw)
                    else:
                        self.scalar.add_spider_pair(self.phase(v), self.phase(w), self.get_params(v), self.get_params(w))
                else:
                    if et == EdgeType.SIMPLE:
                        self.scalar.add_spider_pair(self.phase(v), self.phase(w), self.get_params(v), self.get_params(w))
                    else:
                        self.scalar.add_node(self.phase(v)+self.phase(w), params_vw)
        self.remove_vertices(rem)

    def vdata_dict(self, vertex: VT) -> Dict[str, Any]:
        return { key: self.vdata(vertex, key) for key in self.vdata_keys(vertex) }

    def set_vdata_dict(self, vertex: VT, d: Dict[str, Any]) -> None:
        self.clear_vdata(vertex)
        for k, v in d.items():
            self.set_vdata(vertex, k, v)

    def edata_dict(self, edge: ET) -> Dict[str, Any]:
        """Return a dict of all edge data for the given edge."""
        return { key: self.edata(edge, key) for key in self.edata_keys(edge) }

    def set_edata_dict(self, edge: ET, d: Dict[str, Any]) -> None:
        """Set all edge data for the given edge from a dict, clearing existing data first."""
        self.clear_edata(edge)
        for k, v in d.items():
            self.set_edata(edge, k, v)

    def is_well_formed(self) -> bool:
        """Returns whether the graph is a well-formed ZX-diagram.
        This means that it has no isolated boundary vertices,
        each boundary vertex has a unique neighbor,
        W_input vertices have two neighbors: W_output and other,
        and W_output vertices have at least two neighbors: W_input and other."""
        for v in self.vertices():
            if self.type(v) == VertexType.BOUNDARY:
                if self.vertex_degree(v) != 1:
                    return False
            elif self.type(v) == VertexType.W_INPUT:
                if self.vertex_degree(v) != 2:
                    return False
            elif self.type(v) == VertexType.W_OUTPUT:
                if self.vertex_degree(v) < 2:
                    return False
        return True

    def get_auto_simplify(self) -> bool:
        """Returns whether this graph auto-simplifies parallel edges
        
        For multigraphs, this parameter might change, but simple graphs should always return True."""
        return True

    def set_auto_simplify(self, s: bool) -> None:
        """Set whether this graph auto-simplifies parallel edges
        
        Simple graphs should always auto-simplify, so this method is a no-op."""
        pass
    
    def is_phase_gadget(self, v: VT) -> bool:
        """Returns True if the vertex is the 'hub' of a phase gadget"""
        if not vertex_is_zx(self.type(v)) or self.phase(v) != 0 or self.vertex_degree(v) < 2:
            return False
        for w in self.neighbors(v):
            if vertex_is_zx(self.type(w)) and self.vertex_degree(w) == 1:
                return True
        return False






from fractions import Fraction
import itertools
import math

from .utils import EdgeType, VertexType, toggle_edge
from .linalg import Mat2, Z2
from .simplify import id_simp, tcount
from .rewrite_rules.rules import apply_rule, pivot, match_spider_parallel, spider
from .circuit import Circuit
from .circuit.gates import Gate, ParityPhase, CNOT, HAD, ZPhase, CZ, InitAncilla
from .graph.base import BaseGraph, VT, ET

from typing import List, Optional, Tuple, Dict, Set, Union

from .extract import bi_adj, connectivity_from_biadj, max_overlap
# , greedy_reduction, find_minimal_sums, xor_rows, column_optimal_swap


# permute the rows of "m" such that as many rows as possible have
# 1s in position given by "pos"
# TODO: this can be done in poly-time with graph matching algo
# (e.g. Hopcroft-Karp)
def ones_in_pos(m, pos):
    winner = -1

    perm = None
    for p in itertools.permutations(range(m.rows())):
        score = sum(m[p[i],pos[i]] for i in range(m.rows()))
        if score > winner:
            winner = score
            perm = p
        if score == m.rows():
            break
    m1 = m.copy()

    if not perm is None:
        for i in range(m.rows()):
            m[perm[i],:] = m1[i,:]

# produce a parity matrix that extracts all of the extractable
# vertices in convenient places and ignores the rest
def compute_row_ops(m):
    ops = Mat2.id(m.rows())
    m1 = m.copy()
    m1.gauss(full_reduce=True,x=ops)
    #extr_rows = [r for r,row in enumerate(m1.data) if sum(row)==1]

    # keep only the rows corresponding to extractable verts
    #ops_rows = []
    ops_rows = [ops.data[r]
                   for r,row in enumerate(m1.data)
                   if sum(row)==1]
    #ops_rows.sort(key=lambda row: sum(row), reverse=False)
    ops_rows.sort(reverse=True)
    #ops_rows = ops_rows[0:2]
    ops = Mat2(ops_rows)


    # for r,row in enumerate(m1.data):
    #     if sum(row)==1:
    #         ops_rows.append(ops.data[r])
    # ops = Mat2(ops_rows)
    # ops = Mat2([ops.data[r]
    #               for r,row in enumerate(m1.data)
    #               if sum(row)==1])
    # list.sort(ops.data, reverse=True)
    
    # find pivot columns
    pivot_cols = []
    ops.copy().gauss(pivot_cols=pivot_cols)
    #ones_in_pos(ops,pivot_cols)

    # augment ops by adding unit vectors for all non-pivot columns
    ops2 = Mat2.id(m.rows())
    for r in range(ops.rows()):
        ops2[pivot_cols[r],:] = ops[r,:]
    return ops2


def alt_extract_circuit(
        g:BaseGraph[VT,ET], 
        optimize_czs:bool=True, 
        optimize_cnots:int=2, 
        quiet:bool=True
        ) -> Circuit:
    """
    Experiments with the extractor. (not ready for primetime :)


    Args:
        g: The ZX-diagram graph to be extracted into a Circuit.
        optimize_czs: Whether to try to optimize the CZ-subcircuits by exploiting overlap between the CZ gates
        optimize_cnots: (0,1,2,3) Level of CNOT optimization to apply.
        quiet: Whether to print detailed output of the extraction process.
    """
    qs = g.qubits() # We are assuming that these are objects that update...
    rs = g.rows()   # ...to reflect changes to the graph, so that when...
    # ty = g.types()  # ... g.set_row/g.set_qubit is called, these things update directly to reflect that
    phases = g.phases()
    c = Circuit(g.qubit_count())

    gadgets = {}
    inputs = g.inputs()
    outputs = g.outputs()
    for v in g.vertices():
        if g.vertex_degree(v) == 1 and v not in inputs and v not in outputs:
            n = list(g.neighbors(v))[0]
            gadgets[n] = v
    
    qubit_map: Dict[VT,int] = dict()
    frontier = []
    for i,o in enumerate(outputs):
        v = list(g.neighbors(o))[0]
        if v in inputs: continue
        frontier.append(v)
        qubit_map[v] = i
        
    czs_saved = 0
    q: Union[float,int]
    
    while True:
        # preprocessing

        for v in frontier: # First removing single qubit gates
            q = qubit_map[v]
            b = [w for w in g.neighbors(v) if w in outputs][0]
            e = g.edge(v,b)
            if g.edge_type(e) == 2: # Hadamard edge
                c.add_gate("HAD",q)
                g.set_edge_type(e, EdgeType.SIMPLE)
            if phases[v]: 
                c.add_gate("ZPhase", q, phases[v])
                g.set_phase(v,0)

        # And now on to CZ gates
        cz_mat = Mat2([[0 for i in range(g.qubit_count())] for j in range(g.qubit_count())])
        for v in frontier:
            for w in list(g.neighbors(v)):
                if w in frontier:
                    cz_mat.data[qubit_map[v]][qubit_map[w]] = 1
                    cz_mat.data[qubit_map[w]][qubit_map[v]] = 1
                    g.remove_edge(g.edge(v,w))

        if optimize_czs:
            overlap_data = max_overlap(cz_mat)
            while len(overlap_data[1]) > 2: #there are enough common qubits to be worth optimizing
                i,j = overlap_data[0][0], overlap_data[0][1]
                czs_saved += len(overlap_data[1])-2
                c.add_gate("CNOT",i,j)
                for qb in overlap_data[1]:
                    c.add_gate("CZ",j,qb)
                    cz_mat.data[i][qb]=0
                    cz_mat.data[j][qb]=0
                    cz_mat.data[qb][i]=0
                    cz_mat.data[qb][j]=0
                c.add_gate("CNOT",i,j)
                overlap_data = max_overlap(cz_mat)

        for i in range(g.qubit_count()):
            for j in range(i+1,g.qubit_count()):
                if cz_mat.data[i][j]==1:
                    c.add_gate("CZ",i,j)
        
        # Now we can proceed with the actual extraction
        # First make sure that frontier is connected in correct way to inputs
        neighbor_set = set()
        for v in frontier.copy():
            d = [w for w in g.neighbors(v) if w not in outputs]
            if any(w in inputs for w in d): #frontier vertex v is connected to an input
                if len(d) == 1: # Only connected to input, remove from frontier
                    frontier.remove(v)
                    continue
                # We disconnect v from the input b via a new spider
                b = [w for w in d if w in inputs][0]
                q = qs[b]
                r = rs[b]
                w = g.add_vertex(VertexType.Z, q, r+1)
                e = g.edge(v,b)
                et = g.edge_type(e)
                g.remove_edge(e)
                g.add_edge((v, w), EdgeType.HADAMARD)
                g.add_edge((w,b),toggle_edge(et))
                d.remove(b)
                d.append(w)
            neighbor_set.update(d)
        
        if not frontier: break # No more vertices to be processed. We are done.
        
        # First we check if there is a phase gadget in the way
        removed_gadget = False
        for w in neighbor_set:
            if w not in gadgets: continue
            for v in g.neighbors(w):
                if v in frontier:
                    apply_rule(g,pivot,[(w,v,[],[o for o in g.neighbors(v) if o in g.outputs])]) # type: ignore
                    frontier.remove(v)
                    del gadgets[w]
                    frontier.append(w)
                    qubit_map[w] = qubit_map[v]
                    removed_gadget = True
                    break
        if removed_gadget: # There was indeed a gadget in the way. Go back to the top
            continue
            
        neighbors = list(neighbor_set)
        m = bi_adj(g,neighbors,frontier)
        ops = compute_row_ops(m)
        m1 = ops * m

        cnots = Circuit(g.qubit_count())
        blocksize = math.ceil(math.log(g.qubit_count(),2)) * 2
        winner = -1
        for bs in range(1,blocksize):
            cnots1 = Circuit(g.qubit_count())
            ops.copy().gauss(full_reduce=True,
                y=cnots1, blocksize=bs)
            if winner == -1 or len(cnots1.gates) < winner:
                cnots = cnots1

        #return m, ops

        connectivity_from_biadj(g,m1,neighbors,frontier)

        good_verts = dict()
        for i, row in enumerate(m1.data):
            if sum(row) == 1:
                v = frontier[i]
                w = neighbors[[j for j in range(len(row)) if row[j]][0]]
                good_verts[v] = w
        
        # if not quiet: print("good_verts:", good_verts)
        if not good_verts: #raise Exception("No extractable vertex found. Something went wrong")
            print("No extractable vertex found. Something went wrong")
            break
        hads = []
        for v,w in good_verts.items(): # Update frontier vertices
            hads.append(qubit_map[v])
            #c.add_gate("HAD",qubit_map[v])
            qubit_map[w] = qubit_map[v]
            b = [o for o in g.neighbors(v) if o in outputs][0]
            g.remove_vertex(v)
            g.add_edge((w,b))
            frontier.remove(v)
            frontier.append(w)
        if not quiet: print("Vertices extracted:", len(good_verts))
        #for cnot in cnots: c.add_gate(cnot)
        c.add_circuit(cnots)
        for h in hads: c.add_gate("HAD",h)
    return c
    # TODO: finalise
